from pathlib import PurePosixPath
import pytest
import shutil

from datalad_next.consts import PRE_INIT_COMMIT_SHA
from datalad_next.utils import rmtree

from ..gitdiff import (
    GitTreeItemType,
    GitDiffStatus,
    iter_gitdiff,
)


def test_iter_gitdiff_invalid():
    with pytest.raises(ValueError):
        # no meaningful comparison
        list(iter_gitdiff('.', None, None))
    with pytest.raises(ValueError):
        # unsupported eval mode
        list(iter_gitdiff('.', None, None, eval_submodule_state='weird'))


def test_iter_gitdiff_basic(existing_dataset, no_result_rendering):
    ds = existing_dataset
    dsp = ds.pathobj
    # we compare based on the last state of the corresponding
    # branch if there is any, or the HEAD of the current
    # branch
    comp_base = ds.repo.get_corresponding_branch() or 'HEAD'
    # we use two distinct content blobs below, hardcode sha here
    # for readability
    empty_sha = 'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391'
    content = '123'
    content_sha = 'd800886d9c86731ae5c4a62b0b77c437015e00d2'
    status_args = (
        # we always test against the root of the dataset
        dsp,
        comp_base,
        # we always compare to the worktree
        None,
    )
    diff_args = (
        # we always test against the root of the dataset
        dsp,
        # we always compare to last committed state
        f'{comp_base}~1', comp_base,
    )
    # clean dataset, no items
    assert list(iter_gitdiff(*status_args)) == []
    testpath = dsp / 'sub' / 'test'
    testpath.parent.mkdir()
    testpath.touch()
    # dataset with untracked file, no items
    assert list(iter_gitdiff(*status_args)) == []
    ds.save(to_git=True)
    # clean dataset again, no items
    assert list(iter_gitdiff(*status_args)) == []
    # added file
    diff = list(iter_gitdiff(*diff_args))
    assert len(diff) == 1
    di = diff[0]
    assert di.status == GitDiffStatus.addition
    assert di.name == 'sub/test'
    assert di.prev_name is di.prev_gitsha is di.prev_gittype is None
    assert di.gitsha == empty_sha
    assert di.gittype == GitTreeItemType.file
    # modified file
    testpath.write_text(content)
    diff = list(iter_gitdiff(*status_args))
    assert len(diff) == 1
    di = diff[0]
    # labeled as modified
    assert di.status == GitDiffStatus.modification
    # the name is plain str in POSIX
    assert di.name == di.prev_name == 'sub/test'
    # path conversion yield POSIX relpath
    assert di.path == di.prev_path == PurePosixPath(testpath.relative_to(dsp))
    # unstaged modification reports no shasum
    assert di.gitsha is None
    assert di.prev_gitsha == empty_sha
    assert di.gittype == di.prev_gittype == GitTreeItemType.file
    # make clean
    ds.save(to_git=True)
    moved_testpath = testpath.parent / 'moved_test'
    testpath.rename(moved_testpath)
    # renamed file, unstaged, reported as deletion, we do not see the addition
    # yet (untracked)
    diff = list(iter_gitdiff(*status_args))
    assert len(diff) == 1
    di = diff[0]
    assert di.status == GitDiffStatus.deletion
    assert di.name == di.prev_name == 'sub/test'
    assert di.prev_gitsha == content_sha
    assert di.prev_gittype == GitTreeItemType.file
    assert di.gitsha is di.gittype is None
    # make clean
    ds.save(to_git=True)
    # now we can look at the rename
    diff = list(iter_gitdiff(*diff_args, find_renames=100))
    assert len(diff) == 1
    di = diff[0]
    assert di.status == GitDiffStatus.rename
    assert di.name == 'sub/moved_test'
    assert di.prev_name == 'sub/test'
    assert di.gitsha == di.prev_gitsha == content_sha
    assert di.prev_gittype is di.gittype is GitTreeItemType.file
    assert di.percentage == 100
    # now a copy
    shutil.copyfile(moved_testpath, testpath)
    ds.save(to_git=True)
    diff = list(iter_gitdiff(*diff_args, find_copies=100))
    assert len(diff) == 1
    di = diff[0]
    assert di.status == GitDiffStatus.copy
    assert di.name == 'sub/test'
    assert di.prev_name == 'sub/moved_test'
    assert di.gitsha == di.prev_gitsha == content_sha
    assert di.percentage == 100
    # now replace file with submodule
    testpath.unlink()
    # we must safe to appease datalad's content collision detection
    ds.save(to_git=True)
    # intermediate smoke test for describing a single tree (diff from parents)
    diff = list(iter_gitdiff(dsp, None, comp_base))
    assert len(diff) == 1
    assert diff[0].status == GitDiffStatus.deletion
    # now cause typechange
    ds.create(testpath)
    diff = list(iter_gitdiff(
        dsp,
        # because we have an intermediate safe, compare to two states
        # back
        f'{comp_base}~2', comp_base,
    ))
    assert len(diff) == 2
    # let's ignore the uninteresting .gitmodules addition for further tests
    di = [i for i in diff if i.name != '.gitmodules'][0]
    assert di.status == GitDiffStatus.typechange
    assert di.name == di.prev_name == 'sub/test'
    assert di.gitsha != di.prev_gitsha
    assert di.prev_gitsha == content_sha
    assert di.prev_gittype == GitTreeItemType.file
    assert di.gittype == GitTreeItemType.submodule


def test_iter_gitdiff_nonroot(existing_dataset, no_result_rendering):
    ds = existing_dataset
    comp_base = ds.repo.get_corresponding_branch() or 'HEAD'
    # all tests are concerned with running not in the dataset root
    root = ds.pathobj
    nonroot = root / 'sub'
    nonroot.mkdir()
    status_args = (nonroot, comp_base, None)
    diff_args = (nonroot, f'{comp_base}~1', comp_base)

    # nothing to report, no problem
    assert list(iter_gitdiff(*status_args)) == []
    # change above CWD is not reported
    (root / 'rootfile').touch()
    ds.save(to_git=True)
    assert list(iter_gitdiff(*diff_args)) == []
    # check worktree modification detection too
    (root / 'rootfile').write_text('some')
    assert list(iter_gitdiff(*status_args)) == []
    # and now test that reporting is relative to
    # CWD
    (nonroot / 'nonrootfile').touch()
    ds.save(to_git=True)
    assert list(iter_gitdiff(*diff_args))[0].name == 'nonrootfile'
    (nonroot / 'nonrootfile').write_text('other')
    assert list(iter_gitdiff(*diff_args))[0].name == 'nonrootfile'


def test_iter_gitdiff_nonrec(existing_dataset, no_result_rendering):
    ds = existing_dataset
    dsp = ds.pathobj
    comp_base = ds.repo.get_corresponding_branch() or 'HEAD'
    subdir = dsp / 'sub'
    subdir.mkdir()
    for fn in ('f1.txt', 'f2.txt'):
        (subdir / fn).touch()
    ds.save(to_git=True)
    diff = list(iter_gitdiff(dsp, f'{comp_base}~1', comp_base, recursive='no'))
    assert len(diff) == 1
    di = diff[0]
    assert di.name == 'sub'
    assert di.gittype == GitTreeItemType.directory
    assert di.status == GitDiffStatus.addition
    di_tree = di
    # same behavior for a worktree modification
    for fn in ('f1.txt', 'f2.txt'):
        (subdir / fn).write_text('modified')
    diff = list(iter_gitdiff(dsp, f'{comp_base}~1', None, recursive='no'))
    assert len(diff) == 1
    di = diff[0]
    # these are identical to the diff-tree based report
    for p in ('name', 'gittype', 'prev_gitsha', 'prev_gittype'):
        assert getattr(di, p) == getattr(di_tree, p)
    # and there are different
    # not staged, no gitsha
    assert di.gitsha is None
    # it does no type inference for the previous state (expensive)
    assert di.prev_gittype is None

    # when the directory existed in the from-state it becomes a
    # modification
    diff = list(iter_gitdiff(dsp, f'{comp_base}~1', None, recursive='no'))
    assert len(diff) == 1
    diff[0].status == GitDiffStatus.modification

    # now remove the subdir
    rmtree(subdir)
    diff = list(iter_gitdiff(dsp, comp_base, None, recursive='no'))
    assert len(diff) == 1
    # it still reports a modification, even though the directory is empty/gone.
    # it would require a filesystem STAT to detect a deletion, and a further
    # type investigation in `from_treeish` to detect a type change.
    # This is not done until there is evidence for a real use case
    diff[0].status == GitDiffStatus.modification


def test_iter_gitdiff_typechange_issue6791(
        existing_dataset, no_result_rendering):
    # verify that we can handle to problem described in
    # https://github.com/datalad/datalad/issues/6791
    #
    # a subdataset is wiped out (uncommitted) and replaced by a file
    ds = existing_dataset
    ds.create('subds')
    rmtree(ds.pathobj / 'subds')
    (ds.pathobj / 'subds').touch()
    diff = list(iter_gitdiff(
        ds.pathobj,
        ds.repo.get_corresponding_branch() or 'HEAD', None,
    ))
    assert len(diff) == 1
    di = diff[0]
    assert di.status == GitDiffStatus.typechange
    assert di.name == di.prev_name == 'subds'
    # unstaged change
    assert di.gitsha is None
    assert di.prev_gittype == GitTreeItemType.submodule
    assert di.gittype == GitTreeItemType.file


def test_iter_gitdiff_rec(existing_dataset, no_result_rendering):
    ds = existing_dataset
    subds = ds.create('subds')
    dsp = ds.pathobj
    comp_base = ds.repo.get_corresponding_branch() or 'HEAD'
    status_args = (dsp, comp_base, None)
    diff_args = (dsp, f'{comp_base}~1', comp_base)

    diff = list(iter_gitdiff(*diff_args, recursive='submodules'))
    # we get more than just .gitmodules and a submodule record
    assert len(diff) > 2
    # the entire submodule is new and the first one, so everything
    # is an addition
    assert all(i.status == GitDiffStatus.addition for i in diff)
    # only files, no submodule record, by default
    assert all(i.gittype == GitTreeItemType.file for i in diff)

    # when we ask for it, we get the submodule item too
    diff_w_sm = list(iter_gitdiff(*diff_args,
                                  recursive='submodules',
                                  yield_tree_items='submodules'))
    assert len(diff) + 1 == len(diff_w_sm)
    assert any(i.name == 'subds' and i.gittype == GitTreeItemType.submodule
               for i in diff_w_sm)

    # smoke test for an all-clean diff against the worktrees
    assert list(iter_gitdiff(*status_args, recursive='submodules')) == []

    # make subdataset record modified
    (subds.pathobj / 'file').touch()
    subds.save(to_git=True)
    diff = list(iter_gitdiff(*status_args, recursive='submodules'))
    assert len(diff) == 1
    di = diff[0]
    assert di.name == 'subds/file'
    assert di.status == GitDiffStatus.addition
    # now with submodule item
    diff_w_sm = list(iter_gitdiff(*status_args,
                                  recursive='submodules',
                                  yield_tree_items='all'))
    assert len(diff_w_sm) == 2
    di = diff_w_sm[0]
    # the submodule item is always first
    assert di.name == 'subds'
    assert di.gittype == GitTreeItemType.submodule
    assert di.status == GitDiffStatus.modification
    assert diff_w_sm[1] == diff[0]

    # safe the whole hierarchy
    ds.save(recursive=True)
    # we get the exact same change report via the diff to HEAD~1:HEAD
    assert diff == list(iter_gitdiff(*diff_args, recursive='submodules'))

    # modify a tracked file in the subdataset
    (subds.pathobj / 'file').write_text('123')
    diff_w_sm = list(iter_gitdiff(*status_args,
                                  recursive='submodules',
                                  yield_tree_items='all'))
    # same report for the submodule (and it is first again)
    assert diff_w_sm[0].name == 'subds'
    assert diff_w_sm[0].gittype == GitTreeItemType.submodule
    assert diff_w_sm[0].status == GitDiffStatus.modification
    # but this time the file is not an addition but a modification
    assert diff_w_sm[1].name == 'subds/file'
    assert diff_w_sm[1].status == GitDiffStatus.modification

    # force-wipe the subdataset, and create a condition where the subdatasets
    # is expected but missing
    rmtree(subds.pathobj)
    diff = list(iter_gitdiff(*status_args))
    assert len(diff) == 1
    di = diff[0]
    assert di.name == 'subds'
    assert di.status == GitDiffStatus.deletion
    # if we now run with recursion, we get the exact same result, the absent
    # submodule is a subtree that we do not recurse into, hence the report
    # is only on the tree itself
    assert diff == list(iter_gitdiff(*status_args, recursive='submodules'))
    # use the opportunity to check equality of recursive='all' for this case
    assert diff == list(iter_gitdiff(*status_args, recursive='all'))


def test_iter_gitdiff_multilvl_rec(existing_dataset, no_result_rendering):
    ds = existing_dataset
    s1 = ds.create('sublvl1')
    s1.create('sublvl2')
    ds.save()
    dsp = ds.pathobj

    common_args = dict(
        path=dsp,
        from_treeish=PRE_INIT_COMMIT_SHA,
        to_treeish=ds.repo.get_corresponding_branch() or 'HEAD',
    )
    diff = list(iter_gitdiff(
        # check that we get full repo content from all submodules
        recursive='submodules',
        # check that the container item flags are passed into the
        # recursion properly
        yield_tree_items='submodules',
        **common_args
    ))
    for repo in ('sublvl1', 'sublvl1/sublvl2'):
        assert any(
            d.name == repo and d.gittype == GitTreeItemType.submodule
            for d in diff)
    for base in ('', 'sublvl1/', 'sublvl1/sublvl2/'):
        assert any(
            d.name == f'{base}.datalad/config' \
            and d.gittype == GitTreeItemType.file
            for d in diff)

    # try with very simple pathspec constraint, where the pathspec
    # itself does not match the submodules that contain the
    # matches
    diff = list(iter_gitdiff(
        pathspecs=[':(glob)**/config'],
        recursive='submodules',
        **common_args
    ))
    assert len(diff) == 3
    assert all(
        d.name.endswith('config') and d.gittype == GitTreeItemType.file
        for d in diff
    )



import pytest

from datalad.api import next_status

from datalad_next.constraints import (
    CommandParametrizationError,
    ParameterConstraintContext,
)
from datalad_next.utils import chpwd

from ..status import (
    opt_eval_subdataset_state_values,
    opt_recursive_values,
    opt_untracked_values,
)


def test_status_invalid(tmp_path, datalad_cfg):
    # we want exhaustive parameter validation (i.e., continue after
    # first failure), saves some code here
    datalad_cfg.set('datalad.runtime.parameter-violation',
                    'raise-at-end',
                    scope='global')
    with chpwd(tmp_path):
        with pytest.raises(CommandParametrizationError) as e:
            next_status(
                untracked='weird',
                recursive='upsidedown',
                eval_subdataset_state='moonphase',
            )
        errors = e.value.errors
        assert 'no dataset found' in \
            errors[ParameterConstraintContext(('dataset',))].msg.casefold()
        for opt in ('untracked', 'recursive', 'eval_subdataset_state'):
            assert 'is not one of' in \
                errors[ParameterConstraintContext((opt,))].msg.casefold()


def test_status_renderer_smoke(existing_dataset):
    ds = existing_dataset
    assert ds.next_status() == []
    (ds.pathobj / 'untracked').touch()
    st = ds.next_status()
    assert len(st) == 1


def test_status_clean(existing_dataset, no_result_rendering):
    ds = existing_dataset
    ds.create('subds')
    for recmode in opt_recursive_values:
        assert [] == ds.next_status(recursive=recmode)
    for untracked in opt_untracked_values:
        assert [] == ds.next_status(untracked=untracked)
    for eval_sm in opt_eval_subdataset_state_values:
        assert [] == ds.next_status(eval_subdataset_state=eval_sm)



"""
"""

__docformat__ = 'restructuredtext'

from logging import getLogger
from pathlib import (
    Path,
    PurePosixPath,
)
from urllib.parse import urlparse

import datalad
from datalad_next.commands import (
    EnsureCommandParameterization,
    ValidatedInterface,
    Parameter,
    build_doc,
    datasetmethod,
    eval_results,
    get_status_dict,
)
from datalad_next.exceptions import (
    CapturedException,
    UrlOperationsRemoteError,
)
from datalad_next.utils import ensure_list
from datalad_next.constraints import (
    AnyOf,
    EnsureChoice,
    EnsureDataset,
    EnsureGeneratorFromFileLike,
    EnsureJSON,
    EnsureListOf,
    EnsureMapping,
    EnsurePath,
    EnsureURL,
    EnsureValue,
    WithDescription,
)
from datalad_next.url_operations import AnyUrlOperations

lgr = getLogger('datalad.local.download')


class EnsureURLFilenamePairFromURL(EnsureURL):
    """Accept a URL and derive filename from it path component

    Return original URL and filename as a mapping
    """
    def __init__(self):
        super().__init__(required=['scheme', 'path'])

    def __call__(self, value):
        url = super().__call__(value)
        # derive a filename from the URL.
        # we take the last element of the 'path' component
        # of the URL, or fail
        parsed = urlparse(url)
        filename = PurePosixPath(parsed.path).name
        return {url: filename}


@build_doc
class Download(ValidatedInterface):
    """Download from URLs

    This command is the front-end to an extensible framework for performing
    downloads from a variety of URL schemes. Built-in support for the schemes
    'http', 'https', 'file', and 'ssh' is provided. Extension packages
    may add additional support.

    In contrast to other downloader tools, this command integrates with the
    DataLad credential management and is able to auto-discover credentials.
    If no credential is available, it automatically prompts for them, and
    offers to store them for reuse after a successful authentication.

    Simultaneous hashing (checksumming) of downloaded content is supported
    with user-specified algorithms.

    The command can process any number of downloads (serially). it can read
    download specifications from (command line) arguments, files, or STDIN.
    It can deposit downloads to individual files, or stream to STDOUT.

    Implementation and extensibility

    Each URL scheme is processed by a dedicated handler. Additional
    schemes can be supported by sub-classing
    :class:`datalad_next.url_operations.UrlOperations` and implementing
    the `download()` method. Extension packages can register new handlers,
    by patching them into the `datalad_next.download._urlscheme_handlers`
    registry dict.
    """
    #
    # argument format specifications
    #
    # any URL that we would take must have a scheme, because we switch
    # protocol handling based on that. It is also crucial for distinguishing
    # stuff like local paths and file names from URLs
    url_constraint = EnsureURL(required=['scheme'])
    # other than a plain URL we take a mapping from a URL to a local path.
    # The special value '-' is used to indicate stdout
    # if given as a single string, we support single-space-delimited items:
    # "<url> <path>"
    url2path_constraint = WithDescription(
        EnsureMapping(
            key=url_constraint,
            value=EnsureValue('-') | EnsurePath(),
            delimiter=' ',
            # we disallow length-2 sequences to be able to distinguish from
            # a length-2 list of URLs.
            # the key issue is the flexibility of EnsurePath -- pretty much
            # anything could be a valid unix path
            allow_length2_sequence=False,
        ),
        error_message=f'not a dict, length-2-iterable, or space-delimited str',
    )
    # each specification items is either a mapping url->path, just a url, or a
    # JSON-encoded url->path mapping.  the order is complex-to-simple for the
    # first two (to be able to distinguish a mapping from an encoded URL. The
    # JSON-encoding is tried last, it want match accidentally)
    urlonly_item_constraint = WithDescription(
        EnsureURLFilenamePairFromURL() & url2path_constraint,
        error_message='not a URL with a path component '
        'from which a filename can be derived',
    )
    json_item_constraint = WithDescription(
        EnsureJSON() & url2path_constraint,
        error_message='not a JSON-encoded str with an object or length-2-array',
    )
    any_item_constraint = WithDescription(
        AnyOf(
            # TODO explain
            url2path_constraint,
            urlonly_item_constraint,
            json_item_constraint,
        ),
        error_message='not a single item\n{__itemized_causes__}',
    )

    # we support reading specification items (matching any format defined
    # above) as
    # - a single item
    # - as a list of items
    # - a list given in a file, or via stdin (or any file-like in Python)
    spec_constraint = WithDescription(
        AnyOf(
            any_item_constraint,
            WithDescription(
                EnsureListOf(any_item_constraint),
                error_message='not a list of any such item',
            ),
            WithDescription(
                EnsureGeneratorFromFileLike(
                    any_item_constraint,
                    exc_mode='yield',
                ),
                error_message="not a path to a file with one such item per-line, "
                "nor '-' to read any such item from STDIN",
            ),
        ),
        error_message="does not provide URL->(PATH|-) mapping(s)\n{__itemized_causes__}"
    )

    force_choices = EnsureChoice('overwrite-existing')

    # Interface.validate_args() will inspect this dict for the presence of a
    # validator for particular parameters
    _validator_ = EnsureCommandParameterization(dict(
        spec=spec_constraint,
        # if given, it must also exist as a source for configuration items
        # and/or credentials
        dataset=EnsureDataset(installed=True),
        force=force_choices | EnsureListOf(force_choices),
        # TODO EnsureCredential
        #credential=
        # TODO EnsureHashAlgorithm
        #hash=EnsureHashAlgorithm | EnsureIterableOf(EnsureHashAlgorithm)
    ))

    # this is largely here for documentation and CLI parser building
    _params_ = dict(
        spec=Parameter(
            args=('spec',),
            metavar='<path>|<url>|<url-path-pair>',
            doc="""Download sources and targets can be given in a variety of
            formats: as a URL, or as a URL-path-pair that is mapping a source
            URL to a dedicated download target path. Any number of URLs or
            URL-path-pairs can be provided, either as an argument list, or
            read from a file (one item per line). Such a specification input
            file can be given as a path to an existing file (as a single
            value, not as part of a URL-path-pair). When the special path
            identifier '-' is used, the download is written to STDOUT.
            A specification can also be read in JSON-lines encoding (each line
            being a string with a URL or an object mapping a URL-string to a
            path-string).  [PY: In addition, specifications can also be given
            as a list or URLs, or as a list of dicts with a URL to
            path mapping. Paths are supported in string form, or as `Path`
            objects. PY]""",
            nargs='+'),
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""Dataset to be used as a configuration source. Beyond
            reading configuration items, this command does not interact with
            the dataset."""),
        force=Parameter(
            args=("--force",),
            action='append',
            # TODO only here because datalad-core CLI generates docs from this
            choices=force_choices._allowed,
            doc="""By default, a target path for a download must not exist yet.
            'force-overwrite' disabled this check."""),
        credential=Parameter(
            args=("--credential",),
            metavar='NAME',
            doc="""name of a credential to be used for authorization. If no
            credential is identified, the last-used credential for the
            authentication realm associated with the download target will
            be used. If there is no credential available yet, it will be
            prompted for. Once used successfully, a prompt for entering
            to save such a new credential will be presented.""",
        ),
        hash=Parameter(
            args=("--hash",),
            action='append',
            metavar='ALGORITHM',
            doc="""Name of a hashing algorithm supported by the Python
            'hashlib' module, e.g. 'md5' or 'sha256'.
            [CMD: This option can be given more than once CMD]
            """),
    )

    _examples_ = [
        {'text': 'Download webpage to "myfile.txt"',
         'code_cmd': 'datalad download "http://example.com myfile.txt"',
         'code_py': 'download({"http://example.com": "myfile.txt"})'},
        {'text': 'Read download specification from STDIN (e.g. JSON-lines)',
         'code_cmd': 'datalad download -',
         'code_py': 'download("-")'},
        {'text':
         'Simultaneously hash download, hexdigest reported in result record',
         'code_cmd':
         'datalad download --hash sha256 http://example.com/data.xml"',
         'code_py':
         'download("http://example.com/data.xml", hash=["sha256"])'},
        {'text': 'Download from SSH server',
         'code_cmd': 'datalad download "ssh://example.com/home/user/data.xml"',
         'code_py': 'download("ssh://example.com/home/user/data.xml")'},
        {'text': 'Stream a download to STDOUT',
         'code_cmd': 'datalad -f disabled download "http://example.com -"'},
    ]

    @staticmethod
    @datasetmethod(name="download")
    @eval_results
    def __call__(spec, *, dataset=None, force=None, credential=None,
                 hash=None):
        # which config to inspect for credentials etc
        cfg = dataset.ds.config if dataset else datalad.cfg

        if isinstance(spec, (str, dict)):
            # input validation allows for a non-list item, turn into
            # list for uniform processing
            spec = [spec]

        # cache of already used handlers
        url_handler = AnyUrlOperations(cfg=cfg)

        # we are not running any tests upfront on the whole spec,
        # because the spec can be a generator and consume from a
        # long-running source (e.g. via stdin)
        for item in spec:
            if isinstance(item, CapturedException):
                # the generator encountered an exception for a particular
                # item and is relaying it as per instructions
                # exc_mode='yield'. We report and move on. Outside
                # flow logic will decide if processing continues
                yield get_status_dict(
                    action='download',
                    status='impossible',
                    exception=item,
                )
                continue

            url, dest = item.popitem()
            # we know that any URL has a scheme
            if not url_handler.is_supported_url(url):
                yield get_status_dict(
                    action='download',
                    status='error',
                    message='unsupported URL '
                    '(custom URL handlers can be declared via DataLad '
                    'configuration)',
                    url=url,
                )
                continue

            # ready destination path
            try:
                dest = _prep_dest_path(dest, force)
            except ValueError as e:
                yield get_status_dict(
                    action='download',
                    status='error',
                    exception=CapturedException(e),
                    url=url,
                    path=dest,
                )
                continue

            try:
                download_props = url_handler.download(
                    url,
                    dest,
                    credential=credential,
                    hash=ensure_list(hash),
                )
                res = get_status_dict(
                    action='download',
                    status='ok',
                    url=url,
                    path=dest,
                )
                # take the reported download properties (e.g. any computed
                # hashes a a starting point, and overwrite any potentially
                # conflicting keys with the standard ones)
                res = dict(
                    download_props,
                    **res)
                yield res
            except Exception as e:
                ce = CapturedException(e)
                res = get_status_dict(
                    action='download',
                    status='error',
                    message='download failure',
                    url=url,
                    path=dest,
                    exception=ce,
                )
                if issubclass(type(e), UrlOperationsRemoteError):
                    res['status_code'] = e.status_code
                yield res


def _prep_dest_path(dest, force):
    if dest == '-':
        # nothing to prep for stdout
        return
    # make sure we can replace any existing target path later
    # on. but do not remove here, we might not actually be
    # able to download for other reasons
    if _lexists(dest) and (
            not force or 'overwrite-existing' not in force):
        raise ValueError('target path already exists')

    # create parent directory if needed
    dest.parent.mkdir(parents=True, exist_ok=True)
    return dest


def _lexists(path: Path):
    try:
        path.lstat()
        return True
    except FileNotFoundError:
        return False



import pytest

from ..annexkey import AnnexKey
from ..archivist import ArchivistLocator
from ..enums import ArchiveType

some_key = 'MD5-s389--e9f624eb778e6f945771c543b6e9c7b2'


def test_archivistlocator():
    test_locator = \
        f'dl+archive:{some_key}#path=dir/file.csv&size=234&atype=tar'

    al = ArchivistLocator.from_str(test_locator)

    assert al.akey == AnnexKey.from_str(some_key)
    assert al.atype == ArchiveType.tar

    # round trip
    assert str(al) == test_locator

    # type determination from key
    assert ArchivistLocator.from_str(
        'dl+archive:MD5E-s1--e9f624eb778e6f945771c543b6e9c7b2.tar#path=f.txt'
    ).atype == ArchiveType.tar
    assert ArchivistLocator.from_str(
        'dl+archive:MD5E-s1--e9f624eb778e6f945771c543b6e9c7b2.tgz#path=f.txt'
    ).atype == ArchiveType.tar
    assert ArchivistLocator.from_str(
        'dl+archive:MD5E-s1--e9f624eb778e6f945771c543b6e9c7b2.tar.gz#path=f.txt'
    ).atype == ArchiveType.tar
    assert ArchivistLocator.from_str(
        'dl+archive:MD5E-s1--e9f624eb778e6f945771c543b6e9c7b2.zip#path=f.txt'
    ).atype == ArchiveType.zip


def test_archivistlocatori_errors():
    for wrong in (
        # no chance without prefix
        'bogus',
        # not just a prefix or some bogus properties
        'dl+archive:',
        'dl+archive:#',
        'dl+archive:keything',
        'dl+archive:#props',
        'dl+archive:keything#props',
        # a real key is required, but not sufficient
        f'dl+archive:{some_key}#props',
        # we require a member path, the rest is optional
        f'dl+archive:{some_key}#size=123',
        f'dl+archive:{some_key}#size=123&atype=tar',
        # must be a proper POSIX path, relative, no ..
        f'dl+archive:{some_key}#path=/dummy',
        f'dl+archive:{some_key}#path=../dd',
        # cannot work with unknown archive type
        f'dl+archive:{some_key}#path=good&size=123&atype=eh!',
    ):
        with pytest.raises(ValueError):
            ArchivistLocator.from_str(wrong)



"""Build a complete (minimal) command that implements batch-mode

But without any batch-mode code inside the command implementation
"""

from io import StringIO

from datalad_next.commands import (
    EnsureCommandParameterization,
    ValidatedInterface,
    Parameter,
    build_doc,
    eval_results,
    get_status_dict,
)
from datalad_next.exceptions import CapturedException
from datalad_next.constraints import (
    EnsureGeneratorFromFileLike,
    EnsureJSON,
)


@build_doc
class DoBatch(ValidatedInterface):
    """Explainer!"""
    _validator_ = EnsureCommandParameterization(dict(
        # TODO add constraint that checks composition
        # of each JSON-line
        source=EnsureGeneratorFromFileLike(
            EnsureJSON(),
            exc_mode='yield',
        ),
    ))

    _params_ = dict(
        source=Parameter(args=('source',)),
    )

    @staticmethod
    @eval_results
    def __call__(source):
        for item in source:
            if isinstance(item, CapturedException):
                yield get_status_dict(
                    action='dobatch',
                    status='error',
                    exception=item,
                )
                continue
            yield get_status_dict(
                action='dobatch',
                status='ok',
                selected=item.get('this'),
            )


def test_dobatch(monkeypatch, no_result_rendering):
    data_in = '{"this":[1,2,3],"noise":"some"}\n{"this":true}'
    monkeypatch.setattr('sys.stdin', StringIO(data_in))
    res = DoBatch.__call__('-')
    assert len(res) == 2
    assert res[0]['selected'] == [1, 2, 3]
    assert res[1]['selected'] is True

    # now we have an intermediate error
    monkeypatch.setattr('sys.stdin', StringIO('bug\n' + data_in))
    res = DoBatch.__call__(
        '-', on_failure='ignore')
    assert len(res) == 3
    assert res[0]['status'] == 'error'
    assert 'Expecting value' in res[0]['error_message']
    # second one has the data now
    assert res[1]['selected'] == [1, 2, 3]
    assert res[2]['selected'] is True



"""Constraints for Git-related concepts and parameters"""
from __future__ import annotations

from datalad_next.runners import (
    CommandError,
    call_git,
    call_git_oneline,
)
from .base import (
    Constraint,
    DatasetParameter,
)


class EnsureGitRefName(Constraint):
    """Ensures that a reference name is well formed

    Validation is performed by calling `git check-ref-format`.
    """
    def __init__(self,
                 allow_onelevel: bool = True,
                 normalize: bool = True,
                 refspec_pattern: bool = False):
        """
        Parameters
        ----------
        allow_onelevel:
          Flag whether one-level refnames are accepted, e.g. just 'main'
          instead of 'refs/heads/main'.
        normalize:
          Flag whether a normalized refname is validated and return.
          This includes removing any leading slash (/) characters and
          collapsing runs of adjacent slashes between name components
          into a single slash.
        refspec_pattern:
          Flag whether to interpret a value as a reference name pattern
          for a refspec (allowed to contain a single '*').
        """
        super().__init__()
        self._allow_onelevel = allow_onelevel
        self._normalize = normalize
        self._refspec_pattern = refspec_pattern

    def __call__(self, value: str) -> str:
        if not value:
            # simple, do here
            self.raise_for(value, 'refname must not be empty')

        cmd = ['check-ref-format']
        cmd.append('--allow-onelevel'
                   if self._allow_onelevel
                   else '--no-allow-onelevel')
        if self._refspec_pattern:
            cmd.append('--refspec-pattern')
        if self._normalize:
            cmd.append('--normalize')

        cmd.append(value)

        try:
            res = (call_git_oneline
                   if self._normalize else call_git)(cmd)
        except CommandError as e:
            self.raise_for(
                value,
                'is not a valid refname',
                __caused_by__=e,
            )

        if self._normalize:
            return res
        else:
            return value

    def short_description(self):
        return '{}Git refname{}'.format(
            '(single-level) ' if self._allow_onelevel else '',
            ' or refspec pattern' if self._refspec_pattern else '',
        )


class EnsureRemoteName(Constraint):
    """Ensures a valid remote name, and optionally if such a remote is known
    """
    _label = 'remote'

    def __init__(self,
                 known: bool | None = None,
                 dsarg: DatasetParameter | None = None):
        """
        Parameters
        ----------
        known: bool, optional
           By default, a given value is only checked if it is a syntactically
           correct remote name.
           If ``True``, also checks that the given name corresponds to a
           known remote in the dataset given by ``dsarg``. If ``False``,
           checks that the given remote does not match any known remote
           in that dataset.
        dsarg: DatasetParameter, optional
           Identifies a dataset for testing remote existence, if requested.
        """
        self._label = 'remote'
        self._known = known
        self._dsarg = dsarg

    def __call__(self, value: str) -> str:
        if not value:
            # simple, do here
            self.raise_for(
                value,
                f'missing {self._label} name',
            )

        if self._known is not None:
            assert self._dsarg, \
                f"Existence check for {self._label} requires dataset " \
                "specification"

        # second conditional only to reassert type
        if self._known and self._dsarg is not None:
            # we don't need to check much, only if a remote of this name
            # already exists -- no need to check for syntax compliance
            # again
            if not any(
                k.startswith(f"remote.{value}.")
                for k in self._dsarg.ds.config.keys()
            ):
                self.raise_for(
                    value,
                    f'is not a known {self._label}',
                )
        else:
            # whether or not the remote must not exist, or we would not care,
            # in all cases we need to check for syntax compliance
            EnsureGitRefName(
                allow_onelevel=True,
                refspec_pattern=False,
            )(value)

        if self._known is None:
            # we only need to know that something was provided,
            # no further check
            return value

        assert self._dsarg
        if self._known is False and any(
            k.startswith(f"remote.{value}.")
            for k in self._dsarg.ds.config.keys()
        ):
            self.raise_for(
                value,
                f'name conflicts with a known {self._label}',
            )

        return value

    def short_description(self):
        return f"Name of a{{desc}} {self._label}".format(
            desc=' known' if self._known
            else ' not-yet-known' if self._known is False else ''
        )

    def for_dataset(self, dataset: DatasetParameter) -> Constraint:
        """Return an similarly parametrized variant that checks remote names
        against a given dataset (argument)"""
        return self.__class__(
            known=self._known,
            dsarg=dataset,
        )


class EnsureSiblingName(EnsureRemoteName):
    """Identical to ``EnsureRemoteName``, but used the term "sibling"

    Only error messages and documentation differ, with "remote" being
    replaced with "sibling".
    """
    _label = 'sibling'



"""Connect ``log_progress``-style progress reporting to git-annex, add `close()`

This patch introduces a dedicated progress log handler as a proxy between
standard datalad progress logging and a git-annex special remote as
an approach to report (data transfer) progress to a git-annex parent process.

This functionality is only (to be) used in dedicated special remote processes.

This patch also adds a standard `close()` handler to special remotes, and calls
that handler in a context manager to ensure releasing any resources. This
replaces the custom `stop()` method, which is undocumented and only used by the
`datalad-archive` special remote.

This patch also adds code that allows to patch a class that is already loaded
"""

from contextlib import closing
import logging
from typing import (
    Dict,
    Type,
)

from . import apply_patch
from datalad_next.annexremotes import SpecialRemote


def only_progress_logrecords(record: logging.LogRecord) -> bool:
    """Log filter to ignore any non-progress log message"""
    return hasattr(record, 'dlm_progress')


class AnnexProgressLogHandler(logging.Handler):
    """Log handler to funnel progress logs to git-annex

    For this purpose the handler wraps
    :class:`datalad_next.annexremotes.SpecialRemote` instance.
    When it receives progress log messages, it converts any
    increment reports to absolute values, and then calls
    the special remote's ``send_progress()`` method, which will
    cause the respective progress update protocol message to
    be issued.

    .. note::

       Git-annex only supports "context-free" progress reporting.  When a
       progress report is send, it is assumed to be on a currently running
       transfer. Only a single integer value can be reported, and it
       corresponds to the number of bytes transferred.

       This approach implemented here cannot distinguish progress reports
       that corresponding to git-annex triggered data transfers and other
       (potentially co-occurring) operations. The likelihood of unrelated
       operations reporting progress is relatively low, because this
       handler is only supposed to be used in dedicated special remote
       processes, but remains possible.

       This implementation is set up to support tracking multiple
       processes, and could report one of them selectively. However, at
       present any progress update is relayed to git-annex directly.
       This could lead to confusing and non-linear progress reporting.
    """
    def __init__(self, annexremote: SpecialRemote):
        super().__init__()
        self.annexremote = annexremote
        self._ptrackers: Dict[str, int] = {}

    def emit(self, record: logging.LogRecord):
        """Process a log record

        Any incoming log record, compliant with
        http://docs.datalad.org/design/progress_reporting.html
        is processed. Increment reports are converted to absolute
        values, and each update is eventually passed on to special remote,
        which issues a progress report to git-annex.
        """
        if not hasattr(record, 'dlm_progress'):
            # a filter should have been used to prevent this call
            return

        maint = getattr(record, 'dlm_progress_maint', None)
        if maint in ('clear', 'refresh'):
            return
        pid = getattr(record, 'dlm_progress')
        update = getattr(record, 'dlm_progress_update', None)
        if pid not in self._ptrackers:
            # this is new
            prg = getattr(record, 'dlm_progress_initial', 0)
            self._ptrackers[pid] = prg
            self.annexremote.send_progress(prg)
        elif update is None:
            # not an update -> done
            self._ptrackers.pop(pid)
        else:
            prg = self._ptrackers[pid]
            if getattr(record, 'dlm_progress_increment', False):
                prg += update
            else:
                prg = update
            self._ptrackers[pid] = prg
            self.annexremote.send_progress(prg)


def patched_underscore_main(args: list, cls: Type[SpecialRemote]):
    """Full replacement for datalad.customremotes.main._main()

    Its only purpose is to create a running instance of a SpecialRemote.
    The only difference to the original in datalad-core is that once this
    instance exists, it is linked to a log handler that converts incoming
    progress log messages to the equivalent annex protocol progress reports.

    This additional log handler is a strict addition to the log handling
    setup established at this point. There should be no interference with
    any other log message processing.

    .. seealso::

       :class:`AnnexProgressLogHandler`
    """
    assert cls is not None
    from annexremote import Master

    # Reload the class, to allow `cls` itself to be patched.
    new_module = __import__(cls.__module__, fromlist=[cls.__name__])
    cls = getattr(new_module, cls.__name__)

    master = Master()
    # this context manager use relies on patching in a close() below
    with closing(cls(master)) as remote:
        master.LinkRemote(remote)

        # we add an additional handler to the logger to deal with
        # progress reports
        dlroot_lgr = logging.getLogger('datalad')
        phandler = AnnexProgressLogHandler(remote)
        phandler.addFilter(only_progress_logrecords)
        dlroot_lgr.addHandler(phandler)

        # run the remote
        master.Listen()
        # cleanup special case datalad-core `archive` remote
        # nobody should do this, use `close()`
        if hasattr(remote, 'stop'):
            remote.stop()


# a default cleanup handler for CoreBaseSpecialRemote
# this enables us to use a standard `closing()` context manager with
# special remotes
def specialremote_defaultclose_noop(self):
    pass


apply_patch(
    'datalad.customremotes', 'SpecialRemote', 'close',
    specialremote_defaultclose_noop,
    msg='Retrofit `SpecialRemote` with a `close()` handler',
    expect_attr_present=False,
)
apply_patch(
    'datalad.customremotes.main', None, '_main',
    patched_underscore_main,
    msg='Replace special remote _main() '
    "with datalad-next's progress logging enabled variant")



from datalad_next.tests.marker import skipif_no_network


@skipif_no_network
def test_uncurl_progress_reporting_to_annex(existing_dataset, monkeypatch):
    """Set up a repo that is used to download a key,
    check that we see progress reports
    """
    repo = existing_dataset.repo
    # enable uncurl to get a datalad code piece generate progress info
    repo.call_annex([
        'initremote',
        'uncurl',
        'type=external',
        'externaltype=uncurl',
        'encryption=none',
    ])
    # 1.7MB download, should never change
    testfilekey = 'MD5E-s1725572--3f9f0f5c05517686c008115a611586b1.zip'
    testfileurl = \
        'https://github.com/datalad/datalad/archive/refs/tags/0.18.3.zip'
    testfilename = 'datalad.zip'

    # register the key in the dataset with the source URL
    repo.call_annex(['registerurl', testfilekey, testfileurl])

    # record the key to be available from uncurl
    uncurl_uuid = repo.call_annex_records(['info', 'uncurl'])[0]['uuid']
    repo.call_annex(['setpresentkey', testfilekey, uncurl_uuid, '1'])

    # place key in worktree (not strictly required, but a more common setup)
    repo.call_annex(['fromkey', '--force', testfilekey, testfilename])

    # intercept progress logs in this process.  in order for progress reports
    # to appear here, uncurl needs to report them to git-annex, and our runner
    # setup needs to catch them and call `log_progress`. So this really is an
    # end-to-end test.

    logs = []

    # patch the log_progress() used in annexrepo.py
    def catch_progress(*args, **kwargs):
        logs.append(kwargs)

    import datalad.support.annexrepo
    monkeypatch.setattr(
        datalad.support.annexrepo,
        "log_progress",
        catch_progress,
    )

    # trigger a download. use git-annex directly such that there is
    # little chance that the uncurl remote process is talking to a
    # datalad parent process directly
    repo._call_annex_records(
        args=['get'],
        files=[testfilename],
        progress=True,
        total_nbytes=1725572,
    )
    # check that we got the progress init report
    assert any('total' in log for log in logs)
    # and at least one progress update -- do not check for more, because
    # on fast systems this may take very little time
    assert any('update' in log for log in logs)



#!/usr/bin/env python
## emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""git-remote-datalad-annex to fetch/push via any git-annex special remote

In essence, this Git remote helper bootstraps a utility repository in order
to push/fetch the state of a repository to any location accessible by any
git-annex special remote implementation. All information necessary for this
bootstrapping is taken from the remote URL specification. The internal utility
repository is removed again after every invocation. Therefore changes to the
remote access configuration can be made any time by simply modifying the
configured remote URL.

When installed, this remote helper is invoked for any "URLs" that start with
the prefix ``datalad-annex::``. Following this prefix, two types of
specifications are support.

1. Plain parameters list::

    datalad-annex::?type=<special-remote-type>&[...][exporttree=yes]

   In this case the prefix is followed by a URL query string that comprises all
   necessary (and optional) parameters that would be normally given to the
   ``git annex initremote`` command. It is required to specify the special
   remote ``type``, and it is possible to request "export" mode for any special
   remote that supports it. Depending on the chosen special remote additional
   parameters may be required or supported. Please consult the git-annex
   documentation at https://git-annex.branchable.com/special_remotes/

2. URL::

    datalad-annex::<url>[?...]

   Alternatively, an actual URL can be given after the prefix. In this case,
   the, now optional, URL query string can still be used to specify arbitrary
   parameters for special remote initialization. In addition, the query string
   specification can use Python-format-style placeholder to reference
   particular URL components as parameters values, in order to avoid
   double-specification.

   The list of supported placeholders is ``scheme``, ``netloc``, ``path``,
   ``fragment``, ``username``, ``password``, ``hostname``, ``port``,
   corresponding to the respective URL components. In addition, a ``noquery``
   placeholder is supported, which resolves to the entire URL except any
   query string. An example of such a URL specification is::

    datalad-annex::file:///tmp/example?type=directory&directory={path}&encryption=none'

   which would initialize a ``type=directory`` special remote pointing
   at ``/tmp/example``.


Caution with collaborative workflows

There is no protection against simultaneous, conflicting repository state
uploads from two different locations! Similar to git-annex's "export"
feature, this feature is most appropriately used as a dataset deposition
mechanism, where uploads are conducted from a single site only -- deposited
for consumption by any number of parties.

If this Git remote helper is to be used for multi-way collaboration, with two
or more parties contributing updates, it is advisable to employ a separate
``datalad-annex::`` target site for each contributor, such that only one site
is pushing to any given location. Updates are exchanged by the remaining
contributors adding the respective other ``datalad-annex::`` sites as
additional Git remotes, analog to forks of a repository.


Special remote type support

In addition to the regular list of special remotes, plain http(s) access
via URLs is also supported via the 'web' special remote. For such cases,
only the base URL and the 'type=web' parameter needs to be given, e.g::

    git clone 'datalad-annex::https://example.com?type=web&url={noquery}'

When a plain URL is given, with no parameter specification in a query
string, the parameters ``type=web`` and ``exporttree=yes`` are added
automatically by default. This means that this remote helper can clone
from any remote deposit accessible via ``http(s)`` that matches the layout
depicted in the next section.


Remote layout

The representation of a repository at a remote depends on the chosen type of
special remote. In general, two files will be deposited. One text file
containing a list of Git ``refs`` contained in the deposit, and one ZIP file
with a (compressed) archive of a bare Git repository. Beside the idiosyncrasies
of particular special remotes, to major modes determine the layout of a remote
deposit. In "normal" mode, two annex keys (``XDLRA--refs``,
``XDLRA--repo-export``) will be deposited. In "export" mode, a directory tree is
created that is designed to blend with arbitrary repository content, such that
a git remote and a git-annex export can be pushed to the same location without
conflicting with each other. The aforementioned files will be represented like
this::

    .datalad
    └── dotgit  # named to not be confused with an actual Git repository
        ├── refs
        └── repo.zip

The default LZMA-compression of the ZIP file (in both export and normal mode)
can be turned off with the ``dladotgit=uncompressed`` URL parameter.


Credential handling

Some git-annex special remotes require the specification of credentials via
environment variables. With the URL parameter ``dlacredential=<name>`` it
is possible to query DataLad for a user/password credential to be used for
this purpose. This convenience functionality is supported for the special
remotes ``glacier``, ``s3``, and ``webdav``.

When a credential of the given name does not exist, or no credential name
was specified, an attempt is made to determine a suitable credential based
on, for example, a detected HTTP authentication realm. If no matching
credential could be found, the user will be prompted to enter a credential.
After having successfully established access, the entered credential will
be saved in the local credential store.

DataLad-based credentials are only utilized, when the native git-annex
credential setup via environment variables is not in use (see the documentation
of a particular special remote implementation for more information).


Implementation details

This Git remote implementation uses *two* extra repositories, besides the
repository (R) it is used with, to do its work:

(A) A tiny repository that is entirely bootstrapped from the remote URL,
    and is used to retrieve/deposit a complete state of the actual repo
    an a remote site, via a git-annex special remote setup.

(B) A local, fully functional mirror repo of the remotely stored
    repository state.

On fetch/push the existence of both additional repositories is ensured. The
remote state of retrieved via repo (A), and unpacked to repo (B).  The actual
fetch/push Git operations are performed locally between the repo (R) and
repo (B). On push, repo (B) is then packed up again, and deposited on the
remote site via git-annex transfer in repo (A).

Due to a limitation of this implementation, it is possible that when the last
upload step fails, Git nevertheless advances the pushed refs, making it appear
as if the push was completely successful. That being said, Git will still issue
a message (``error: failed to push some refs to..``) and the git-push process
will also exit with a non-zero status. In addition, all of the remote's refs
will be annotated with an additional ref named
``refs/dlra-upload-failed/<remote-name>/<ref-name>`` to indicate the upload
failure. These markers will be automatically removed after the next successful
upload.

.. note::

   Confirmed to work with git-annex version 8.20211123 onwards.

.. todo::

   - At the moment, only one format for repository deposition is supported
     (a ZIP archive of a working bare repository). However this is not
     a good format for the purpose of long-term archiving, because it
     require a functional Git installation to work with. It would be fairly
     doable to make the deposited format configurable, and support additional
     formats. An interesting one would be a fast-export stream, basically
     a plain text serialization of an entire repository.
   - recognize that a different repo is being pushed over an existing
     one at the remote
   - think about adding additional information into the header of `refs`
     maybe give it some kind of stamp that also makes it easier to validate
     by the XDLRA backend
   - think about preventing duplication between the repo and its local
     mirror could they safely share git objects? If so, in which direction?
"""

from __future__ import annotations

__all__ = ['RepoAnnexGitRemote']

import datetime
import logging
import os
import sys
import zipfile
from pathlib import Path
from shutil import make_archive
from typing import (
    IO,
)
from unittest.mock import patch
from urllib.parse import (
    unquote,
    urlparse,
)

from datalad.core.local.repo import repo_from_path

from datalad_next.consts import (
    PRE_INIT_COMMIT_SHA,
    on_windows,
)
from datalad_next.constraints import EnsureInt
from datalad_next.datasets import (
    LegacyAnnexRepo as AnnexRepo,
    LegacyGitRepo as GitRepo,
)
from datalad_next.exceptions import CapturedException
from datalad_next.runners import (
    CommandError,
    call_git,
    call_git_oneline,
    call_git_success,
)
from datalad_next.uis import ui_switcher as ui
from datalad_next.utils import (
    CredentialManager,
    external_versions,
    get_specialremote_credential_envpatch,
    get_specialremote_credential_properties,
    needs_specialremote_credential_envpatch,
    patched_env,
    rmtree,
    specialremote_credential_envmap,
    update_specialremote_credential,
)

lgr = logging.getLogger('datalad.gitremote.datalad_annex')


class RepoAnnexGitRemote(object):
    """git-remote-helper implementation

    ``communicate()`` is the entrypoint.
    """
    # hard code relevant keynames for the XDLRA backend
    # this will always have the refs list
    refs_key = 'XDLRA--refs'
    # this will have the repository archive
    repo_export_key = 'XDLRA--repo-export'
    xdlra_key_locations = {
        refs_key: dict(
            prefix='3f7/4a3', loc='.datalad/dotgit/refs'),
        repo_export_key: dict(
            prefix='eb3/ca0', loc='.datalad/dotgit/repo.zip'),
    }
    # all top-level content in a repo archive
    # this is used as a positive-filter when extracting downloaded
    # archives (to avoid writing to undesirable locations from
    # high-jacked archives)
    safe_content = [
        'branches', 'hooks', 'info', 'objects', 'refs', 'config',
        'packed-refs', 'description', 'HEAD',
    ]
    # define all supported options, including their type-checker
    support_githelper_options = {
        'verbosity': EnsureInt(),
    }
    # supported parameters that can come in via the URL, but must not
    # be relayed to `git annex initremote`
    internal_parameters = ('dladotgit=uncompressed', 'dlacredential=')

    def __init__(
        self,
        gitdir: str,
        remote: str,
        url: str,
        instream: IO = sys.stdin,
        outstream: IO = sys.stdout,
        errstream: IO = sys.stderr,
    ):
        """
        Parameters
        ----------
        gitdir : str
          Path to the GITDIR of the repository to operate on (provided
          by Git).
        remote : str
          Remote label to use (provided by Git).
        url : str
          URL of the remote (provided by Git).
        instream :
          Stream to read communication from Git from.
        outstream :
          Stream to communicate outcomes to Git.
        errstream :
          Stream for logging.
        """
        self.repo = GitRepo(gitdir)
        # this is the key piece, take special remote params from
        # URL
        # this function yields a type= parameter in any case
        self.initremote_params = get_initremote_params_from_url(url)
        self.remote_name = remote
        # internal logic relies on workdir to be an absolute path
        self.workdir = Path(gitdir, 'dl-repoannex', remote).resolve()
        self._repoannexdir = self.workdir / 'repoannex'
        if self._repoannexdir.exists():
            # whatever existed here before is an undesirable
            # leftover of a previous crash
            rmtree(str(self._repoannexdir), ignore_errors=True)
        self._repoannex = None
        self._mirrorrepodir = self.workdir / 'mirrorrepo'
        self._mirrorrepo = None

        # cache for remote refs, to avoid repeated queries
        self._cached_remote_refs: None | str = None

        self.instream = instream
        self.outstream = outstream
        self.errstream = errstream

        # options communicated by Git
        # https://www.git-scm.com/docs/gitremote-helpers#_options
        self.options: dict[str, str] = {}

        # we want to go for verbose output whenever datalad's log level is
        # debug or even more verbose. This makes it unnecessary to call
        # git directly with multiple `-v` options
        self.verbosity_threshold = 1 if lgr.getEffectiveLevel() > 10 else 10

        # ID of the tree to export, if needed
        self.exporttree: None | str = None

        self.credman = None
        self.pending_credential = None
        # must come after the two above!
        self.credential_env = self._get_credential_env()

        annex_version = external_versions['cmd:annex']
        if annex_version < '8.20211123':
            self.log(
                f'git-annex version {annex_version} is unsupported, '
                'please upgrade',
                level=1
            )

    def _get_credential_env(self) -> dict[str, str] | None:
        """
        Returns
        -------
        dict or None
          A dict with all required items to patch the environment, or None
          if not enough information is available.

        Raises
        ------
        ValueError
          If a credential retrieval is requested for an unsupported special
          remote type.
        """
        credential_name = None
        credential_names = [
            p[14:] for p in self.initremote_params
            if p.startswith('dlacredential=')
        ] or None
        if credential_names:
            credential_name = credential_names[0]
        remote_type = self._get_remote_type()
        supported_remote_type = remote_type in specialremote_credential_envmap
        if credential_name and not supported_remote_type:
            # we have no idea how to deploy credentials for this remote type
            raise ValueError(
                f"Deploying credentials for type={remote_type} special "
                "remote is not supported. Remove dlacredential= parameter from "
                "the remote URL and provide credentials according to the "
                "documentation of this particular special remote.")

        if not needs_specialremote_credential_envpatch(remote_type):
            return None

        cred = self._retrieve_credential(credential_name)

        if not cred:
            lgr.debug(
                'Could not find a matching credential for special remote %s',
                self.initremote_params)
            return None

        return get_specialremote_credential_envpatch(remote_type, cred)

    def _retrieve_credential(
            self,
            name: str | None,
    ) -> dict[str, str] | None:
        """Retrieve a credential

        Successfully retrieved credentials are also placed in
        self.pending_credential to be picked up by `_store_credential()`.

        Returns
        -------
        dict or None
          If a credential could be retrieved, a dict with 'user' and
          'secret' keys will be return, or None otherwise.
        """
        if not self.credman:
            self.credman = CredentialManager(self.repo.config)
        assert self.credman is not None

        cred = None
        credprops: dict[str, str] = {}
        if name:
            # we can ask blindly first, caller seems to know what to do
            cred = self.credman.get(
                name=name,
                # give to make legacy credentials accessible
                _type_hint='user_password',
            )
        if not cred:
            # direct lookup failed, try query.
            credprops = get_specialremote_credential_properties(
                self.initremote_params) or {}
            if credprops:
                creds = self.credman.query(_sortby='last-used', **credprops)
                if creds:
                    name, cred = creds[0]
        if not cred:
            # credential query failed too, enable manual entry
            credprops['type'] = 'user_password'
            cred = self.credman.get(
                # this might still be None
                name=name,
                _type_hint='user_password',
                _prompt=f'A credential is required for access',
                # inject anything we already know to make sure we store it
                # at the very end, and can use it for discovery next time
                **credprops
            )

        if not cred:
            return None
        # stage for eventual (re-)storage after having proven to work
        self.pending_credential = (name, cred)
        return {k: cred[k] for k in ('user', 'secret')}

    def _get_remote_type(self) -> str | None:
        remote_type = [
            p[5:] for p in self.initremote_params
            if p.startswith('type=')
        ]
        if not remote_type:
            return None
        return remote_type[0]

    def _store_credential(self) -> None:
        """Look for a pending credential and store it

        Safe to call unconditionally.
        """
        if self.pending_credential and self.credman:
            name, cred = self.pending_credential
            update_specialremote_credential(
                self._get_remote_type(),
                self.credman,
                name,
                cred,
                credtype_hint='user_password',
                duplicate_hint=
                'Specify a credential name via the dlacredential= '
                'remote URL parameter, and/or configure a credential '
                'with the datalad-credentials command{}'.format(
                    f' with a `realm={cred["realm"]}` property'
                    if 'realm' in cred else ''),
            )

    def _ensure_workdir(self) -> None:
        self.workdir.mkdir(parents=True, exist_ok=True)

    @property
    def repoannex(self) -> AnnexRepo:
        """Repo annex repository

        If accessed when there is no repo annex, as new one is created
        automatically. It is bootstrapped entirely from the parameters
        encoded in the remote URL.

        Returns
        -------
        AnnexRepo
          This is always an annex repository. It is configured with
          a single special remote, parameterized from the Git repo URL.

        Raises
        ------
        CommandError
        ValueError
        """
        if self._repoannex:
            return self._repoannex

        self._ensure_workdir()
        try:
            # check if there is one already, would only be due to a prior
            # RUD (rapid unscheduled disassembly)
            ra = repo_from_path(self._repoannexdir)
        except ValueError:
            # funny dance to get to a bare annexrepo
            ra = GitRepo(
                self._repoannexdir,
                create=not GitRepo.is_valid(self._repoannexdir),
                bare=True,
            )
        try:
            # send annex into private mode, if supported
            # this repo will never ever be shared
            call_git_success(['config', 'annex.private', 'true'],
                             cwd=ra.pathobj, capture_output=True)
            call_git_success(['annex', 'init'], capture_output=True)
            ra = AnnexRepo(self._repoannexdir)
            if 'type=web' in self.initremote_params:
                self._init_repoannex_type_web(ra)
            else:
                # let git-annex-initremote take over
                with patched_env(**(self.credential_env or {})):
                    ra.call_annex(
                        ['initremote', 'origin'] + [
                            p for p in self.initremote_params
                            if not any(p.startswith(ip)
                                       for ip in self.internal_parameters)
                        ])
                # make the new remote config known in the repo instance
                ra.config.reload()
            if 'exporttree=yes' in self.initremote_params:
                # conflicts with type=web, but we checked that above already.
                # plant the to-be-exported tree, still leaving the underlying
                # keys unfulfilled
                self.exporttree = make_export_tree(ra)

        except (CommandError, ValueError):
            # something blew up. clean up and blow again
            rmtree(ra.path, ignore_errors=True)
            raise

        self._repoannex = ra
        return ra

    def _init_repoannex_type_web(self, repoannex: AnnexRepo) -> None:
        """Uses registerurl to utilize the omnipresent type=web remote

        Raises
        ------
        ValueError
          When there is no `url=` parameter or when there are other
          parameters than the additional `type=web` and `exporttree=yes`,
          indicating an unsupported setup.
        """
        # for type=web we have to add URLs by hand
        baseurls = [
            v for v in self.initremote_params
            if v.startswith('url=')]
        if not len(baseurls) == 1:
            raise ValueError(
                "'web'-type remote requires 'url' parameter")
        # validate the rest of the params, essentially there
        # must not be any other
        if not all(p in ('type=web', 'exporttree=yes') or p.startswith('url=')
                   for p in self.initremote_params):
            raise ValueError(
                "'web'-type remote only supports 'url' "
                "and 'exporttree' parameters")
        baseurl = baseurls[0][4:]
        for key, kinfo in self.xdlra_key_locations.items():
            repoannex.call_annex([
                'registerurl',
                key,
                f'{baseurl}/{kinfo["loc"]}'
                if 'exporttree=yes' in self.initremote_params else
                f'{baseurl}/{kinfo["prefix"]}/{key}/{key}'
            ])

    @property
    def mirrorrepo(self) -> GitRepo:
        """Local remote mirror repository

        If accessed when there is no local mirror repo, as new one
        is created automatically, either from the remote state (if there is
        any), or an empty one.

        Returns
        -------
        GitRepo
          This is always only a plain Git repository (bare).
        """
        if self._mirrorrepo:
            return self._mirrorrepo

        # ensure we have a mirror repo, either fresh or existing
        self._ensure_workdir()
        if not self.get_remote_refs():
            existing_repo = False
            # there is nothing at the remote, hence we must wipe
            # out the local state, whatever it was to make git
            # report subsequent pushes properly, and prevent
            # "impossible" fetches
            if self._mirrorrepodir.exists():
                # if we extract, we cannot tollerate left-overs
                rmtree(str(self._mirrorrepodir), ignore_errors=True)
                # null the repohandle to be reconstructed later on-demand
                self._mirrorrepo = None
        elif GitRepo.is_valid(self._mirrorrepodir):
            # so we have remote refs and we also have a local mirror
            # create an instance, assume it is set up how we need it
            # must also have bare=True, or the newly created one below
            # will inherit the config
            # https://github.com/datalad/datalad/issues/6347
            mr = GitRepo(self._mirrorrepodir, bare=True)
            # make sure any recursion back in here is prevented
            self._mirrorrepo = mr
            # this will trigger a download if possible (remote has refs)
            self.replace_mirrorrepo_from_remote_deposit_if_needed()
            # reevaluate
            existing_repo = GitRepo.is_valid(self._mirrorrepodir)
        else:
            # we have nothing local, pull from the remote, because it
            # reports stuff to exist
            self.replace_mirrorrepo_from_remote_deposit()
            existing_repo = True

        # (re-)create an instance
        mr = GitRepo(
            self._mirrorrepodir,
            # if the remote had no refs, there would still be no repo
            create=not existing_repo,
            bare=True)

        if not existing_repo:
            # align HEAD symbolic ref between source and mirror repo
            # IF WE CREATED IT LOCALLY JUST NOW, otherwise take whatever
            # we got.
            # otherwise we can end up in a conflict situation where the mirror
            # points to 'master' (or something else) and the source actually
            # has 'main' (or something different)
            src_head_ref = call_git_oneline(
                ['symbolic-ref', 'HEAD'],
                cwd=self.repo.pathobj,
            ).strip()
            call_git_success(
                ['symbolic-ref', 'HEAD', src_head_ref],
                cwd=mr.pathobj,
                capture_output=True,
            )

        self.log('Established mirror')
        self._mirrorrepo = mr
        return mr

    def log(
        self,
        *args,
        level: int = 2,
    ) -> None:
        """Send log messages to the errstream"""
        # A value of 0 for <n> means that processes operate quietly,
        # and the helper produces only error output.
        # 1 is the default level of verbosity,
        # and higher values of <n> correspond to the number of -v flags
        # passed on the command line
        if int(self.options.get('verbosity',
                                self.verbosity_threshold)) >= level:
            print('[DATALAD-ANNEX]', *args, file=self.errstream)

    def send(self, msg: str) -> None:
        """Communicate with Git"""
        print(msg, end='', file=self.outstream, flush=True)

    def communicate(self) -> None:
        """Implement the necessary pieces of the git-remote-helper protocol

        Uses the input, output and error streams configured for the
        class instance.
        """
        self.log('Git remote startup: '
                 f'{self.remote_name} [{self.initremote_params}]')
        for line in self.instream:
            self.log(f'Received Git remote command: {repr(line)}', level=4)
            if line == '\n':
                # orderly exit command
                return
            elif line == 'capabilities\n':
                self.send(
                    'option\n'
                    'list\n'
                    'connect\n'
                    '\n'
                )
            elif line == 'connect git-receive-pack\n':
                # "receive", because we are receiving at the local mirror repo
                # from a `send-pack` process that is connected to the main local
                # repo
                self.log('Connecting git-receive-pack\n')
                self.send('\n')
                # we assume the mirror repo is in-sync with the remote at
                # this point
                pre_refs = sorted(self.mirrorrepo.for_each_ref_(),
                                  key=lambda x: x['refname'])
                # must not capture -- git is talking to it directly from here
                call_git(
                    ['receive-pack', self.mirrorrepo.path],
                    cwd=self.mirrorrepo.pathobj,
                )
                post_refs = sorted(self.mirrorrepo.for_each_ref_(),
                                   key=lambda x: x['refname'])
                if pre_refs != post_refs \
                        or (post_refs != self.get_remote_refs()):
                    # there was a change in the refs of the mirror repo
                    # OR
                    # the mirror is out-of-sync with the remote (could be a
                    # slightly more expensive test)
                    # we must upload it.
                    try:
                        self.replace_remote_deposit_from_mirrorrepo()
                    except Exception:
                        # the bad thing is that we have no way of properly
                        # signaling to git that this happened,
                        # the refs for this remote will look as if the upload
                        # was successful

                        # we do not need to roll-back the refs in the
                        # mirrorrepo as it will be rsync'ed to the remote on
                        # next access
                        self.log('Remote update failed, flagging refs',
                                 post_refs)
                        for ref in post_refs:
                            # best MIH can think of is to leave behind another
                            # ref to indicate the unsuccessful upload
                            call_git_success([
                                'update-ref',
                                # strip 'refs/heads/' from refname
                                f'refs/dlra-upload-failed/{self.remote_name}/'
                                f'{ref["refname"][11:]}',
                                ref['objectname']],
                                cwd=self.repo.pathobj,
                                capture_output=True,
                            )
                        raise

                # clean-up potential upload failure markers for this particular
                # remote. whatever has failed before, we just uploaded a mirror
                # that was freshly sync'ed with the remote state before
                for ref in self.repo.for_each_ref_(
                        fields=('refname',),
                        pattern=f'refs/dlra-upload-failed/{self.remote_name}'):
                    call_git_success(
                        ['update-ref', '-d', ref['refname']],
                        cwd=self.repo.pathobj,
                        capture_output=True,
                    )
                # we do not need to update `self._cached_remote_refs`,
                # because we end the remote-helper process here
                # everything has worked, if we used a credential, update it
                self._store_credential()
                return
            elif line == 'connect git-upload-pack\n':
                # "upload", because we are uploading from the local mirror repo
                # to a `fetch-pack` process that is connected to the main local
                # repo
                try:
                    self.get_remote_refs(raise_on_error=True)
                except Exception as e:
                    self.log("fatal: couldn't find remote refs (repository deposit does not exist, or is inaccessible", level=1)
                    self.log(f"query error: {e!r}", level=2)
                    return

                self.log('Connecting git-upload-pack\n')
                self.send('\n')
                # must not capture -- git is talking to it directly from here.
                # the `self.mirrorrepo` access will ensure that the mirror
                # is up-to-date
                call_git(
                    ['upload-pack', self.mirrorrepo.path],
                    cwd=self.mirrorrepo.pathobj,
                )
                # everything has worked, if we used a credential, update it
                self._store_credential()
                return
            elif line.startswith('option '):
                key, value = line[7:].split(' ', maxsplit=1)
                if key not in self.support_githelper_options:
                    self.send('unsupported\n')
                else:
                    try:
                        self.options[key] = \
                            self.support_githelper_options[key](
                                value.rstrip('\n'))
                        self.send('ok\n')
                    except ValueError as e:
                        # ensure no-multiline message
                        excstr = str(e).replace('\n', '\\n')
                        # git may not communicate reason for error, do log
                        self.log(
                            f'Type-checking of "{line[:-1]}" failed: {excstr}')
                        self.send(f'error {excstr}\n')
            else:
                self.log('UNKNOWN COMMAND', line)
                # unrecoverable error
                return

    def replace_remote_deposit_from_mirrorrepo(self) -> None:
        """Package the local mirrorrepo up, and copy to the special remote

        The mirror is assumed to be ready/complete. It will be cleaned with
        `gc` to minimize the upload size. The mirrorrepo is then compressed
        into an LZMA ZIP archive, and a separate refs list for it is created
        in addition. Both are then copied to the special remote.
        """
        self.log('Replace remote from mirror')
        mirrorrepo = self.mirrorrepo
        repoannex = self.repoannex

        # trim it down, as much as possible
        call_git(['gc'], cwd=mirrorrepo.pathobj)

        # update the repo state keys
        # it is critical to drop the local keys first, otherwise
        # `setkey` below will not replace them with new content
        # however, git-annex fails to do so in some edge cases
        # https://git-annex.branchable.com/bugs/Fails_to_drop_key_on_windows___40__Access_denied__41__/?updated
        # no regular `drop` works, nor does `dropkeys`
        #self.log(repoannex.call_annex(['drop', '--force', '--all']))
        # nuclear option remains, luckily possible in this utility repo
        if on_windows:
            objdir = self.repoannex.dot_git / 'annex' / 'objects'
            if objdir.exists():
                rmtree(str(objdir), ignore_errors=True)
                objdir.mkdir()
        else:
            # more surgical for the rest
            self.log(repoannex.call_annex([
                'dropkey', '--force', self.refs_key, self.repo_export_key]))

        # use our zipfile wrapper to get an LZMA compressed archive
        # via the shutil convenience layer
        with patch('zipfile.ZipFile',
                   UncompressedZipFile
                   if 'dladotgit=uncompressed' in self.initremote_params
                   else LZMAZipFile):
            # TODO exclude hooks (the mirror is always plain-git),
            # would we ever need any
            archive_file = make_archive(
                str(self.workdir / 'repoarchive'),
                'zip',
                root_dir=str(mirrorrepo.path),
                base_dir=os.curdir,
            )
            # hand over archive to annex
            repoannex.call_annex([
                'setkey',
                self.repo_export_key,
                archive_file
            ])
        # generate a list of refs
        # write to file
        refs_file = self.workdir / 'reporefs'
        refs_file.write_text(_format_refs(mirrorrepo))
        self.log(refs_file.read_text())
        # hand over reflist to annex
        self.log(repoannex.call_annex([
            'setkey',
            self.refs_key,
            str(refs_file),
        ]))
        if 'exporttree=yes' in self.initremote_params:
            # we want to "force" an export, because the content of our
            # keys can change, but this is not possible.
            # we cheat be exporting "nothing" (an empty tree) first,
            # and then reexport
            try:
                self.log(repoannex.call_annex(
                    ['export', PRE_INIT_COMMIT_SHA, '--to=origin']))
            except Exception as e:
                # some remotes will error out if we unexport something that
                # wasn't actually exported (e.g. webdav)
                CapturedException(e)
                pass
            self.log(repoannex.call_annex(
                ['export', self.exporttree, '--to=origin']))
        else:
            # it is critical to drop the keys from the remote first, otherwise
            # `copy` below will fail to replace their content
            self.log(repoannex.call_annex(
                ['drop', '--force', '-f', 'origin', '--all']))
            self.log(repoannex.call_annex(
                ['copy', '--fast', '--to', 'origin', '--all']))
        # update remote refs from local ones
        # we just updated the remote from local
        self._cached_remote_refs = self.get_mirror_refs()

    def replace_mirrorrepo_from_remote_deposit_if_needed(
            self,
    ) -> tuple[str | None, str]:
        """Replace the mirror if the remote has refs and they differ
        """
        self.log("Check if mirror needs to be replaced with remote state")
        remote_refs = self.get_remote_refs()
        mirror_refs = self.get_mirror_refs()
        if remote_refs and remote_refs != mirror_refs:
            self.log(repr(remote_refs), repr(mirror_refs))
            # we must replace the local mirror with the
            # state of the remote
            self.replace_mirrorrepo_from_remote_deposit()
        return remote_refs, mirror_refs

    def replace_mirrorrepo_from_remote_deposit(self) -> None:
        """Replaces the local mirror repo with one obtained from the remote

        This method assumes that the remote does have one. This should be
        checked by inspecting `get_remote_refs()` before calling this method.
        """
        self.log('Set mirror to remote state')
        ra = self.repoannex
        # we have to get the key with the repo archive
        # because the local repoannex is likely a freshly bootstrapped one
        # without any remote awareness, claim that the remote has this key
        sremotes = ra.get_special_remotes()
        if len(sremotes) == 1:
            # in case of the 'web' special remote, we have no actual special
            # remote, but URLs for the two individual keys
            ra.call_annex(['setpresentkey', self.repo_export_key,
                           sremotes.popitem()[0], '1'])
        # drop locally to ensure re-downlad, the keyname never changes,
        # even when the content does
        self.log(
            ra.call_annex([
                'drop', '--force',
                '--key', self.repo_export_key])
        )
        # download the repo archive
        self.log(
            ra.call_annex(['get', '--key', self.repo_export_key])
        )
        # locate it in the local annex, use annex function to do this in order
        # to cope with any peculiar repo setups we might face across platforms
        repoexportkeyloc = ra.call_annex_oneline([
            'contentlocation', self.repo_export_key])
        repoexportkeyloc = ra.dot_git / repoexportkeyloc

        if self._mirrorrepodir.exists():
            # if we extract, we cannot tollerate left-overs
            rmtree(str(self._mirrorrepodir), ignore_errors=True)
            # null the repohandle to be reconstructed later on-demand
            self._mirrorrepo = None

        self.log('Extracting repository archive')
        legacy_deposit = False
        with zipfile.ZipFile(repoexportkeyloc) as zip:
            try:
                zip.getinfo('repo/')
                legacy_deposit = True
                safe_content = [f'repo/{i}' for i in self.safe_content]
            except KeyError:
                safe_content = self.safe_content

            zip.extractall(
                self._mirrorrepodir,
                # a bit of a safety-net, exclude all unexpected content
                members=[
                    m for m in zip.namelist()
                    if any(m.startswith(prefix)
                           for prefix in safe_content)],
            )
        if legacy_deposit:
            legacy_basedir = self._mirrorrepodir / 'repo'
            for p in legacy_basedir.iterdir():
                p.rename(self._mirrorrepodir / p.relative_to(legacy_basedir))

    def get_remote_refs(self, raise_on_error: bool = False) -> str | None:
        """Report remote refs

        The underlying special remote is asked whether it has the key
        containing the refs list for the remote. If it does, it is retrieved
        and reported.

        Returns
        -------
        str or None
          If the remote has a refs record, it is returned as a string,
          formatted like a refs file in a Git directory.
          Otherwise, `None` is returned.
        """
        if self._cached_remote_refs:
            # this process already queried them once, return cache
            return self._cached_remote_refs

        self.log("Get refs from remote")
        ra = self.repoannex

        # in case of the 'web' special remote, we have no actual special
        # remote, but URLs for the two individual keys
        sremotes = ra.get_special_remotes()
        # if we do not have a special remote reported, fall back on
        # possibly recorded URLs for the XDLRA--refs key
        sremote_id = sremotes.popitem()[0] if sremotes else 'web'

        # we want to get the latest refs from the remote under all
        # circumstances, and transferkey will not attempt a download for
        # a key that is already present locally -> drop first
        ra.call_annex([
            'drop', '--force', '--key', self.refs_key])
        # now get the key from the determined remote
        try:
            ra.call_annex([
                'transferkey', self.refs_key, f'--from={sremote_id}'])
        except CommandError as e:
            if raise_on_error:
                raise
            CapturedException(e)
            self.log("Remote appears to have no refs")
            # download failed, we have no refs
            # this can happen for legit reasons (prior first push),
            # but also with broken URLs or lack of permissions
            return None

        refskeyloc = ra.call_annex_oneline([
            'contentlocation', self.refs_key])
        # read, cache, return
        refs = (ra.dot_git / refskeyloc).read_text()
        self._cached_remote_refs = refs
        return refs

    def get_mirror_refs(self) -> str:
        """Return the refs of the current mirror repo

        Returns
        -------
        str
        """
        self.log("Get refs from mirror")
        return _format_refs(self.mirrorrepo)


# TODO propose as addition to AnnexRepo
# https://github.com/datalad/datalad/issues/6316
def call_annex_success(self, args, files=None) -> bool:
    """Call git-annex and return true if the call exit code of 0.

    All parameters match those described for `call_annex`.

    Returns
    -------
    bool
    """
    try:
        self.call_annex(args, files)
    except CommandError:
        return False
    return True


class LZMAZipFile(zipfile.ZipFile):
    """Tiny wrapper to monkey-patch zipfile in order to have
    shutil.make_archive produce an LZMA-compressed ZIP"""
    def __init__(self, *args, **kwargs):
        kwargs.pop('compression', None)
        return super().__init__(
            *args, compression=zipfile.ZIP_LZMA, **kwargs)


class UncompressedZipFile(zipfile.ZipFile):
    """Tiny wrapper to monkey-patch zipfile in order to have
    shutil.make_archive produce an uncompressed ZIP"""
    def __init__(self, *args, **kwargs):
        kwargs.pop('compression', None)
        return super().__init__(
            *args, compression=zipfile.ZIP_STORED, **kwargs)


def _format_refs(repo: GitRepo) -> str:
    """Helper to format a standard refs list from for_each_ref() output

    Parameters
    ----------
    repo: GitRepo
      Repo which to query for the 'HEAD' symbolic ref

    Returns
    -------
    str
      Formatted refs list
    """
    refs = repo.for_each_ref_()

    # generate a list of refs
    refstr = '\n'.join(
        "{objectname} {refname}".format(**r)
        for r in refs
    )
    if refstr:
        refstr += '\n'
    refstr += '@{} HEAD\n'.format(
        call_git_oneline(
            ['symbolic-ref', 'HEAD'],
            cwd=repo.pathobj,
        ).strip()
    )
    return refstr


def get_initremote_params_from_url(url: str) -> list[str]:
    """Parse a remote URL for initremote parameters

    Parameters are taken from a URL's query string. In the query
    parameters can be defined directly, or via placeholder
    for all URL components (using Python's format language).

    The following placeholders are supported: 'scheme', 'netloc',
    'path', 'fragment', 'username', 'password', 'hostname',
    'port'. Their values are determined by urlparse(). There is
    no placeholder for the 'query' component, but a 'noquery'
    placeholder is supported, which provides the original
    (reassembled) URL without the query string.

    Parameters
    ----------
    url : str

    Returns
    -------
    list
      git-annex initremote parameter list. Each value string has the format
      'key=value'.
    """
    if url.startswith('datalad-annex::'):
        url = url[15:]
    if not url:
        raise ValueError("Given URL only contained 'datalad-annex::' prefix")
    pu = urlparse(url)
    expansion = {
        p: getattr(pu, p)
        for p in (
            'scheme',
            'netloc',
            'path',
            # we do not extract the 'query', because it is the very
            # thing we iterate over below
            'fragment',
            'username',
            'password',
            'hostname',
            'port')
    }
    expansion['noquery'] = pu._replace(query='').geturl()
    # expand all parameters in the query
    params = [
        # unquote any string -- should be safe, because
        # initremote parameter names should not have any special
        # characters
        unquote(
            # apply any expansion from the URL components
            v.format(**expansion)
        )
        for v in pu.query.split('&')
        # nothing to pull from an empty string
        if v
    ]
    if all(not p.startswith('type=') for p in params):
        # if there is no type declared, this is a plain type=web
        # export using the full URL
        params = ['type=web', 'exporttree=yes', f'url={url}']

    return params


def make_export_tree(repo: GitRepo) -> str:
    """Create an exportable tree

    The function expects a clean (bare) repository. It requires no checkout,
    and does not modify any branches or creates commits.

    The tree is always the same, but still has to be create in the repoannex
    to be accessible for git-annex. It looks like this::

        .datalad
        └── dotgit
            ├── refs
            └── repo.zip

    where the two files under ``dotgit/`` link to the two critical keys. The
    placement of the files under ``.datalad/`` is chosen so that the export can
    blend with an export of the underlying dataset without conflict. The name
    ``dotgit`` rather than ``.git`` is chosen to avoid confusing it with
    an actual nested Git repo.

    Parameters
    ----------
    repo: AnnexRepo
      Repository instance to write to.

    Returns
    -------
    str
        ID of the tree object, suitable for `git-annex export`.
    """
    here = repo.config.get('annex.uuid')
    # reuse existing, or go with fixed random one
    origin = repo.config.get('remote.origin.annex-uuid',
                             '8249ffce-770a-11ec-9578-5f6af5e76eaa')
    assert here, "No 'here'"
    assert origin, "No 'origin'"
    # we need to force Git to use a throwaway index file to maintain
    # the bare nature of the repoannex, git-annex would stop functioning
    # properly otherwise
    index_file = repo.pathobj / 'datalad_tmp_index'
    with patched_env(GIT_INDEX_FILE=index_file):
        try:
            for key, kinfo in RepoAnnexGitRemote.xdlra_key_locations.items():
                # create a blob for the annex link
                linkhash = call_git_oneline(
                    ['hash-object', '-w', '--stdin'],
                    cwd=repo.pathobj,
                    input=f'../../.git/annex/objects/{kinfo["prefix"]}/{key}/{key}',
                ).strip()
                # place link into a tree
                call_git_success(
                    ['update-index', '--add', '--cacheinfo', '120000',
                     linkhash, kinfo["loc"]],
                    cwd=repo.pathobj,
                    capture_output=True,
                )
            # write the complete tree, and return ID
            exporttree = call_git_oneline(
                ['write-tree'], cwd=repo.pathobj
            ).strip()
            # this should always come out identically
            # unless we made changes in the composition of the export tree
            assert exporttree == '7f0e7953e93b4c9920c2bff9534773394f3a5762'

            # clean slate
            if index_file.exists():
                index_file.unlink()
            # fake export.log record
            # <unixepoch>s <here>:<origin> <exporttree>
            now_ts = datetime.datetime.now().timestamp()
            exportlog = call_git_oneline(
                ['hash-object', '-w', '--stdin'],
                input=f'{now_ts}s {here}:{origin} {exporttree}\n',
                cwd=repo.pathobj,
            ).strip()
            call_git_success(
                ['read-tree', 'git-annex'],
                cwd=repo.pathobj,
            )
            call_git_success(
                ['update-index', '--add', '--cacheinfo', '100644',
                 exportlog, 'export.log'],
                cwd=repo.pathobj,
                capture_output=True,
            )
            gaupdate = call_git_oneline(
                ['write-tree'], cwd=repo.pathobj,
            ).strip()
            gacommit = call_git_oneline(
                ['commit-tree', '-m', 'Fake export', '-p', 'git-annex',
                 gaupdate],
                cwd=repo.pathobj,
            ).strip()
            call_git_success(
                ['update-ref', 'refs/heads/git-annex', gacommit],
                cwd=repo.pathobj,
            )
        finally:
            if index_file.exists():
                index_file.unlink()

    return exporttree


def push_caused_change(operations: list[str]) -> bool:
    ok_operations = (
        'new-tag', 'new-branch', 'forced-update', 'fast-forward', 'deleted'
    )
    return any(o in operations for o in ok_operations)


def push_error(operations: list[str]) -> bool:
    error_operations = (
        'no-match', 'rejected', 'remote-rejected', 'remote-failure',
        'error',
    )
    return any(o in operations for o in error_operations)


def main(gitremote_cls=RepoAnnexGitRemote):
    """git-remote helper executable entrypoint"""
    try:
        if len(sys.argv) < 3:
            raise ValueError(f"Usage: {sys.argv[0]} REMOTE-NAME URL")

        remote, url = sys.argv[1:3]
        # provided by Git
        gitdir = os.environ.pop('GIT_DIR')
        # no fallback, must be present
        if gitdir is None:
            raise RuntimeError('GIT_DIR environment variable not defined')

        # stdin/stdout will be used for interactions with git
        # the 'annex' backend really doesn't do much annex-specific
        # albeit maybe progress reporting (unclear to MIH right now)
        # but it does make credential entry possible here, despite the
        # remote helper process being connected to Git with its stdin/stdout
        ui.set_backend('annex')

        # lock and load
        remote = gitremote_cls(gitdir, remote, url)
        remote.communicate()
        # there is no value in keeping around the downloads
        # we either have things in the mirror repo or have to
        # redownload anyways
        # leaving the table clean and always bootstrap from scratch
        # has the advantage that we always automatically react to any
        # git-remote reconfiguration between runs
        rmtree(remote.repoannex.path, ignore_errors=True)
    except Exception as e:
        ce = CapturedException(e)
        # Receiving an exception here is "fatal" by definition.
        # Mimicking git's error reporting style.
        print(f"fatal: {ce}", file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    main()



"""git-annex special remote *archivist* for obtaining files from archives
"""

from __future__ import annotations

# General TODO for future improvements
#
# `datalad.archivist.archive-cache-mode=<name>`
#   Choice of archive (access) caching behavior. ``<name>`` can be any of
#
#   ``persistent-whole``
#     This causes an archive to be downloaded completely on first access to any
#     archive member. A regular ``annex get`` is performed and an archive is
#     placed at its standard location in the local annex. Any archive member
#     will be extracted from this local copy.
#
# Some ideas on optional additional cache modes related to dropping as much as
# possible after the special remote is done. However, these modes also come
# with potential issues re parallel access (what if another remote process
# is still using a particular archive... Think about that when there is a
# real need
#
#  ``keep-downloads``
#    No caching will be performed per se. However, when archive member access
#    happens to require a full archive download, a downloaded archive will
#    not be removed after member extraction. In such cases, this mode will
#    behave like ``persistent-whole``.
#
#  ``none``
#    This is behaving like ``keep-downloads``, but any downloaded archive
#    will be dropped again after extraction is complete.

from collections.abc import Iterable
from dataclasses import dataclass

from pathlib import Path
from shutil import copyfileobj
from typing import (
    Dict,
    Generator,
    List,
    Tuple,
)

from datalad_next.archive_operations import ArchiveOperations

# we intentionally limit ourselves to the most basic interface
# and even that we only need to get a `ConfigManager` instance.
# If that class would support a plain path argument, we could
# avoid it entirely
from datalad_next.datasets import LegacyAnnexRepo

from datalad_next.exceptions import CommandError
from datalad_next.types import (
    AnnexKey,
    ArchivistLocator,
    ArchiveType,
)

from . import (
    RemoteError,
    SpecialRemote,
    UnsupportedRequest,
    super_main
)


class ArchivistRemote(SpecialRemote):
    """git-annex special remote *archivist* for obtaining files from archives

    Successor of the `datalad-archive` special remote. It claims and acts on
    particular archive locator "URLs", registered for individual annex keys
    (see :class:`datalad_next.types.archivist.ArchivistLocator`). These
    locators identify another annex key that represents an archive (e.g., a
    tarball or a zip files) that contains the respective annex key as a member.
    This special remote trigger the extraction of such members from any
    candidate archive when retrieval of a key is requested.

    This special remote cannot store or remove content. The desired usage
    is to register a locator "URL" for any relevant key via
    ``git annex addurl|registerurl`` or ``datalad addurls``.


    Configuration
    -------------

    The behavior of this special remote can be tuned via a number of
    configuration settings.

    `datalad.archivist.legacy-mode=yes|[no]`
      If enabled, all special remote operations fall back onto the
      legacy ``datalad-archives`` special remote implementation. This mode is
      only provided for backward-compatibility. This legacy implementation
      unconditionally downloads archive files completely, and keeps an
      internal cache of the full extracted archive around. The implied
      200% (or more) storage cost overhead for obtaining a complete dataset
      can be prohibitive for datasets tracking large amount of data
      (in archive files).


    Implementation details
    ----------------------

    *CHECKPRESENT*

    When performing a non-download test for the (continued) presence of an
    annex key (as triggered via ``git annex fsck --fast`` or ``git annex
    checkpresentkey``), the underlying archive containing a key will NOT be
    inspected. Instead, only the continued availability of the annex key for
    the containing archive will be tested.  In other words: this implementation
    trust the archive member annotation to be correct/valid, and it also trusts
    the archive content to be unchanged. The latter will be generally the case,
    but may no with URL-style keys.

    Not implementing such a trust-approach *would* have a number of
    consequences.  Depending on where the archive is located (local/remote) and
    what format it is (fsspec-inspectable or not), we would need to download it
    completely in order to verify a matching archive member.  Moreover, an
    archive might also reference another archive as a source, leading to a
    multiplication of transfer demands.
    """
    def __init__(self, annex):
        super().__init__(annex)
        # central archive handler cache, initialized on-prepare
        self._ahandlers = None
        # a potential instance of the legacy datalad-archives implementation
        self._legacy_special_remote = None

    def __getattribute__(self, name: str):
        """Redirect top-level API calls to legacy implementation, if needed"""
        lsr = SpecialRemote.__getattribute__(self, '_legacy_special_remote')
        if lsr is None or name not in (
            'initremote',
            'prepare',
            'claimurl',
            'checkurl',
            'checkpresent',
            'remove',
            'whereis',
            'transfer_retrieve',
            'stop',
        ):
            # we are not in legacy mode or this is no top-level API call
            return SpecialRemote.__getattribute__(self, name)

        return getattr(lsr, name)

    def initremote(self):
        """This method does nothing, because the special remote requires no
        particular setup.
        """
        pass

    def prepare(self):
        """Prepare the special remote for requests by git-annex

        If the special remote is instructed to run in "legacy mode", all
        subsequent operations will be processed by the ``datalad-archives``
        special remote implementation!
        """
        # we have to do this here, because the base class `.repo` will only give
        # us a `LeanAnnexRepo`.
        # TODO it is unclear to MIH what is actually needed API-wise of the legacy
        # interface. Needs research.
        self._repo = LegacyAnnexRepo(self.annex.getgitdir())
        # are we in legacy mode?
        # let remote-specific setting take priority (there could be
        # multiple archivist-type remotes configured), and use unspecific switch
        # as a default, with a general default of NO
        if self.get_remote_gitcfg(
                'archivist', 'legacy-mode', default='no').lower() == 'yes':
            # ATTENTION DEBUGGERS!
            # If we get here, we will bypass all of the archivist
            # implementation! Check __getattribute__() -- pretty much no
            # other code in this file will run!!!
            # __getattribute__ will relay all top-level operations
            # to an instance of the legacy implementation
            from datalad.customremotes.archives import ArchiveAnnexCustomRemote
            lsr = ArchiveAnnexCustomRemote(self.annex)
            lsr.prepare()
            # we can skip everything else, it won't be triggered anymore
            self._legacy_special_remote = lsr
            return

        # central archive key handler coordination
        self._ahandlers = _ArchiveHandlers(
            self.repo,
            # TODO
            #cache_mode=self._getcfg(
            #    'archive-cache-mode',
            #    default='').lower(),
        )

    def claimurl(self, url: str) -> bool:
        """Returns True for :class:`~datalad_next.types.archivist.ArchivistLocator`-style URLs

        Only a lexical check is performed. Any other URL will result in
        ``False`` to be returned.
        """
        try:
            ArchivistLocator.from_str(url)
            return True
        except Exception:
            return False

    def checkurl(self, url: str) -> bool:
        """Parses :class:`~datalad_next.types.archivist.ArchivistLocator`-style URLs

        Returns ``True`` for any syntactically correct URL with all
        required properties.

        The implementation is identical to ``claimurl()``.
        """
        try:
            ArchivistLocator.from_str(url)
        except Exception as e:
            self.message(f'Invalid URL {url!r}: {e}', type='debug')
            return False

        # we should be able to work with this.
        # do not actually test whether the archive is around or whether
        # the path actually points to a member in the archive,
        # leave to transfer_retrieve
        # Do not give detailed info to git-annex for now
        # https://github.com/Lykos153/AnnexRemote/issues/60
        #if member_props.get('size'):
        #    return dict(
        #        filename=member_props['path'].name,
        #        size=member_props['size'],
        #    )
        #else:
        #    return dict(filename=member_props['path'].name)
        return True

    def checkpresent(self, key: str) -> bool:
        """Verifies continued availability of the archive referenced by the key

        No content verification of the archive, or of the particular archive
        member is performed. See "Implementation details" of this class
        for a rational.

        Returns
        -------
        bool
            True if the referenced archive key is present on any remote.
            False if not.
        """
        # the idea here is that: as long as the archive declared to contain
        # the key is still accessible, we declare CHECKPRESENT.
        # In other words: we trust the archive member annotation to be
        # correct/valid.
        # not trusting it would have sever consequences. depending on
        # where the archive is located (local/remote) and what format it
        # is (fsspec-inspectable), we might need to download it completely
        # in order to verify a matching archive member. Moreover, an archive
        # might also reference another archive as a source, leading to a
        # multiplication of transfer demands

        # get all associated archive keys, turn into set because any key might
        # map to multiple archive keys, and we only need to check them once
        akeys = set(
            str(ArchivistLocator.from_str(url).akey)
            for url in self._get_key_dlarchive_urls(key)
        )
        # As with transfer_retrieve blindly checking akeys in arbitrary
        # order is stupid. We should again sort by (local) availability.
        # if we have an archive locally we can check faster, we could check
        # more precisely (actually look into it).
        # We only need to find one archive with a hit, if we search clever
        # we can exit earlier.
        # So let's do a two-pass approach, first check local availability
        # for any archive key, and only if that does not find us an archive
        # go for the remotes
        if any(_get_key_contentpath(self.repo, akey) for akey in akeys):
            # any one is good enough
            # TODO here we could actually look into the archive and
            # verify member presence without relatively little cost
            return True

        for akey in akeys:
            # we leave all checking logic to git-annex
            try:
                # if it exits clean, the key is still present at at least one
                # remote
                self.repo.call_annex(['checkpresentkey', akey])
                return True
            except CommandError:
                self.message(
                    f'Archive key candidate {akey} for key {key} '
                    'not present in any known remote or here',
                    type='debug')

        # when we end up here, we have tried all known archives keys and
        # found none to be present in any known location
        return False

    def transfer_retrieve(self, key: str, localfilename: str):
        """Retrieve an archive member from a (remote) archive

        All registered locators for a requested key will be sorted by
        availability and size of the references archives. For each archive
        the most suitable handler will be initialized, and extraction
        of the identified member will be attempted. If that fails, the next
        handler is tried until all candidate handlers are exhausted.
        Depending on the archive availability and type, archives may need
        to be retrieved from remote sources.
        """
        # rely on from_locators() to bring the candidate archives
        # in some intelligent order to try one after the other.
        # break ASAP to prevent unnecessary processing
        msgs = []
        try:
            for handler, locs in self._ahandlers.from_locators([
                    ArchivistLocator.from_str(url)
                    for url in self._get_key_dlarchive_urls(key)]):
                with Path(localfilename).open('wb') as dst_fp:
                    for loc in locs:
                        try:
                            with handler.open(loc.member) as fp:
                                # TODO progress reporting
                                # but what progress? the extraction
                                # may just be one part, there could also
                                # be file retrieval
                                copyfileobj(fp, dst_fp)
                            return
                        except Exception as e:
                            msg = f'Failed to extract {key!r} from ' \
                                  f'{handler} ({loc.member}): {e}'
                            self.message(msg, type='debug')
                            msgs.append(msg)
        except Exception as e:
            raise RemoteError(f'Could not obtain {key!r}') from e

        raise RemoteError(f'Could not obtain {key!r} from any archive')

    def transfer_store(self, key: str, filename: str):
        """Raises ``UnsupportedRequest``. This operation is not supported."""
        raise UnsupportedRequest('This remote cannot store content')

    def remove(self, key: str):
        """Raises ``UnsupportedRequest``. This operation is not supported."""
        raise UnsupportedRequest('This remote cannot remove content')

    #
    # Helpers
    #

    def _get_key_dlarchive_urls(self, key):
        return self.annex.geturls(key, prefix='dl+archive:')


def main():
    """CLI entry point installed as ``git-annex-remote-archivist``"""
    super_main(
        cls=ArchivistRemote,
        remote_name='archivist',
        description=\
        "access to annex keys stored within other archive-type annex keys ",
    )


#
# Internal helpers
#

@dataclass
class _ArchiveInfo:
    """Representation of an archive used internally by ``_ArchiveHandlers``"""
    local_path: Path | None
    handler: ArchiveOperations | None = None
    type: ArchiveType | None = None


class _ArchiveHandlers:
    """Wraps annex repo to provide access to keys given by ArchivistLocator(s)

    The main functionality is provided by ``from_locators()``.
    """
    # TODO make archive access caching behavior configurable from the outside
    def __init__(self, repo):
        # mapping of archive keys to an info dict
        self._db: Dict[AnnexKey, _ArchiveInfo] = {}
        # for running git-annex queries against the repo
        self._repo = repo

    def from_locators(
            self, locs: List[ArchivistLocator]
    ) -> Generator[Tuple[ArchiveOperations, Iterable[ArchivistLocator]],
                   None, None]:
        """Produce archive handlers for the given locators

        Yield them one-by-one in a maximally intelligent order for efficient
        retrieval (i.e., handlers for archives that are already available
        locally first. Each handlers is yielded fully prepared, i.e.
        if necessary an archive is retrieved before the handler is yielded.
        Therefore a consumer should not fully consume the returned
        generator when an operation can be completed before all handlers
        are exhausted.

        Parameters
        ----------
        locs: List[ArchivistLocator]
          Any number of locators that must all refer to the same annex key
          (key, not archive annex key!).

        Yields
        ------
        ArchiveOperations, Iterable[ArchivistLocator]
          The referenced archive annex keys are de-duplicated and sorted by
          (local) availability and size.  For each archive key a suitable
          ``ArchiveOperations`` handler is yielded together with the locators
          matching the respective archive.
        """
        # determine all candidate source archive keys
        akeys = set(loc.akey for loc in locs)
        # determine which of the known handlers point to a local archive,
        # yield those
        for akey, kh in {
                akey: self._db[akey]
                for akey in akeys
                if akey in self._db and self._db[akey].handler
        }.items():
            # local_path will be None now, if not around
            if kh.local_path:
                # we found one with a local archive.
                # yield handler and all matching locators
                yield kh.handler, [loc for loc in locs if loc.akey == akey]
                # if we get here, this did not work, do not try again
                akeys.remove(akey)
        # of the handlers we do not yet know, which ones have local data,
        # yield those
        for akey in [k for k in akeys if k not in self._db]:
            ainfo = self._get_archive_info(akey, locs)
            # cache for later
            self._db[akey] = ainfo
            if not ainfo.local_path:
                # do not try a local handler, but keep the akey itself in the
                # race, we might need to try "remote" access later on
                continue

            handler = self._get_local_handler(ainfo)
            # store for later
            ainfo.handler = handler
            # yield handler and all matching locators
            yield handler, [loc for loc in locs if loc.akey == akey]
            # if we get here, this did not work, do not try again
            akeys.remove(akey)

        # of the handlers we do know, but do not have local data,
        # possibly obtain the archive, yield those
        #
        # this is the same as the first loop, but this time all local
        # paths are checked, and some akeys might already have been
        # removed
        for akey, kh in {
                akey: self._db[akey]
                for akey in akeys
                if akey in self._db and self._db[akey].handler
        }.items():
            yield handler, [loc for loc in locs if loc.akey == akey]
            # if we get here, this did not work, do not try again
            akeys.remove(akey)

        # all that is left is to create "remote" handlers and yield them.
        # collect any exceptions to report them at the end, if needed
        exc = []
        # but this time sort the keys to start with the smallest ones
        # (just in case a download is involved)
        for akey in sorted(akeys, key=lambda x: x.size):
            # at this point we must have an existing _ArchiveInfo record
            # for this akey
            ainfo = self._db[akey]
            # but we do not have a handler yet
            assert ainfo.handler is None
            try:
                handler = self._get_remote_handler(akey, ainfo)
            except Exception as e:
                exc.append(e)
                continue
            # if this worked, store the handler for later
            ainfo.handler = handler
            yield handler, [loc for loc in locs if loc.akey == akey]

        # if we get here we can stop -- everything was tried. If there were
        # exceptions, make sure to report them
        if exc:
            # TODO better error
            e = RuntimeError(
                'Exhausted all candidate archive handlers '
                f'(previous failures {exc})')
            e.errors = exc
            raise e

    def _get_archive_info(
            self,
            akey: AnnexKey,
            locs: Iterable[ArchivistLocator],
    ) -> _ArchiveInfo:
        # figure out if the archive is local first
        local_path = _get_key_contentpath(self._repo, str(akey))

        # get all reported archive types
        akey_atypes = set(
            loc.atype for loc in locs if loc.akey == akey and loc.atype
        )
        # if we have (consistent) information, pick the type, if not
        # set to None/ignore and wait for type detection by handler
        akey_atype = None if len(akey_atypes) != 1 else akey_atypes.pop()

        ainfo = _ArchiveInfo(
            local_path=local_path,
            type=akey_atype,
        )
        # cache for later
        self._db[akey] = ainfo
        return ainfo

    def _get_local_handler(self, ainfo: _ArchiveInfo) -> ArchiveOperations:
        if not ainfo.type:
            # TODO we could still do mime-type detection. We have the
            # archive file present locally.
            # check datalad-core how it is done in archive support
            raise NotImplementedError

        if ainfo.type == ArchiveType.tar:
            from datalad_next.archive_operations import TarArchiveOperations
            return TarArchiveOperations(
                ainfo.local_path,
                cfg=self._repo.config,
            )
        elif ainfo.type == ArchiveType.zip:
            from datalad_next.archive_operations import ZipArchiveOperations
            return ZipArchiveOperations(
                ainfo.local_path,
                cfg=self._repo.config,
            )
        else:
            raise NotImplementedError

    def _get_remote_handler(
            self,
            akey: AnnexKey,
            ainfo: _ArchiveInfo,
    ) -> ArchiveOperations:
        # right now we have no remote handlers available
        # TODO: use akey to ask the repo for URLs from which the key
        # would be available and select a remote handler to work
        # with that URL
        # instead we retrieve the archive
        res = self._repo.get(str(akey), key=True)
        # if the akey was already around, `res` could be an empty list.
        # however, under these circumstances we should not have ended
        # up here. assert to alert on logic error in that case
        assert isinstance(res, dict)
        if res.pop('success', None) is not True:
            # TODO better error
            raise RuntimeError(f'Failed to download archive key: {res!r}')
        # now we have the akey locally
        ainfo.local_path = _get_key_contentpath(self._repo, str(akey))
        return self._get_local_handler(ainfo)


def _get_key_contentpath(repo: LegacyAnnexRepo, key: str):
    """Return ``Path`` to a locally present annex key, or ``None``

    ``None`` is return when there is not such key present locally.
    """
    try:
        # if it exits clean, there will be a content location
        # and the content can be found at the location
        loc = next(repo.call_annex_items_(['contentlocation', key]))
        # convert to path. git-annex will report a path relative to the
        # dotgit-dir
        # TODO platform-native?
        loc = repo.dot_git / Path(loc)
    except CommandError:
        loc = None
    return loc



"""Custom exceptions raised by ``Constraint`` implementations"""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from textwrap import indent
from types import MappingProxyType
from typing import (
    Any,
    Dict,
    Tuple,
)

# needed for imports in other pieced of the ``constraints`` module
from datalad_next.exceptions import NoDatasetFound


class ConstraintError(ValueError):
    # we derive from ValueError, because it provides the seemingly best fit
    # of any built-in exception. It is defined as:
    #
    #   Raised when an operation or function receives an argument that has
    #   the right type but an inappropriate value, and the situation is not
    #   described by a more precise exception such as IndexError.
    #
    # In general a validation error can also occur because of a TypeError, but
    # ultimately what also matters here is an ability to coerce a given value
    # to a target type/value, but such an exception is not among the built-ins.
    # Moreover, many pieces of existing code do raise ValueError in practice,
    # and we aim to be widely applicable with this specialized class
    """Exception type raised by constraints when their conditions are violated

    A primary purpose of this class is to provide uniform means for
    communicating information on violated constraints.
    """
    def __init__(self,
                 constraint,
                 value: Any,
                 msg: str,
                 ctx: Dict[str, Any] | None = None):
        """
        Parameters
        ----------
        constraint: Constraint
          Instance of the ``Constraint`` class that determined a violation.
        value:
          The value that is in violation of a constraint.
        msg: str
          A message describing the violation. If ``ctx`` is given too, the
          message can contain keyword placeholders in Python's ``format()``
          syntax that will be applied on-access.
        ctx: dict, optional
          Mapping with context information on the violation. This information
          is used to interpolate a message, but may also contain additional
          key-value mappings. A recognized key is ``'__caused_by__'``, with
          a value of one exception (or a list of exceptions) that led to a
          ``ConstraintError`` being raised.
        """
        # the msg/ctx setup is inspired by pydantic
        # we put `msg` in the `.args` container first to match where
        # `ValueError` would have it. Everything else goes after it.
        super().__init__(msg, constraint, value, ctx)

    @property
    def msg(self):
        """Obtain an (interpolated) message on the constraint violation

        The error message template can be interpolated with any information
        available in the error context dict (``ctx``). In addition to the
        information provided by the ``Constraint`` that raised the error,
        the following additional placeholders are provided:

        - ``__value__``: the value reported to have caused the error
        - ``__itemized_causes__``: an indented bullet list str with on
          item for each error in the ``caused_by`` report of the error.

        Message template can use any feature of the Python format mini
        language. For example ``{__value__!r}`` to get a ``repr()``-style
        representation of the offending value.
        """
        msg_tmpl = self.args[0]
        # get interpolation values for message formatting
        # we need a copy, because we need to mutate the dict
        ctx = dict(self.context)
        # support a few standard placeholders
        # the verbatim value that caused the error: with !r and !s both
        # types of stringifications are accessible
        ctx['__value__'] = self.value
        if self.caused_by:
            ctx['__itemized_causes__'] = indent(
                '\n'.join(f'- {str(c)}' for c in self.caused_by),
                "  ",
            )
        return msg_tmpl.format(**ctx)

    @property
    def constraint(self):
        """Get the instance of the constraint that was violated"""
        return self.args[1]

    @property
    def caused_by(self) -> Tuple[Exception] | None:
        """Returns a tuple of any underlying exceptions that caused a violation
        """
        cb = self.context.get('__caused_by__', None)
        if cb is None:
            return None
        elif isinstance(cb, Exception):
            return (cb,)
        else:
            return tuple(cb)

    @property
    def value(self):
        """Get the value that violated the constraint"""
        return self.args[2]

    @property
    def context(self) -> MappingProxyType:
        """Get a constraint violation's context

        This is a mapping of key/value-pairs matching the ``ctx`` constructor
        argument.
        """
        return MappingProxyType(self.args[3] or {})

    def __str__(self):
        return self.msg

    def __repr__(self):
        # rematch constructor arg-order, because we put `msg` first into
        # `.args`
        return '{0}({2!r}, {3!r}, {1!r}, {4!r})'.format(
            self.__class__.__name__,
            *self.args,
        )


class ConstraintErrors(ConstraintError):
    """Exception representing context-specific ConstraintError instances

    This class enables the association of a context in which any particular
    constraint was violated. This is done by passing a mapping, of a context
    identifier (e.g., a label) to the particular ``ConstraintError`` that
    occurred in this context, to the constructor.

    This is a generic implementation with no requirements regarding the
    nature of the context identifiers (expect for being hashable). See
    ``CommandParametrizationError`` for a specialization.
    """
    def __init__(self, exceptions: Dict[Any, ConstraintError]):
        super().__init__(
            # this is the main payload, the base class expects a Constraint
            # but only stores it
            constraint=exceptions,
            # all values are already on record in the respective exceptions
            # no need to pass again
            value=None,
            # no support for a dedicated message here (yet?), pass an empty
            # string to match assumptions
            msg='',
            # and no context
            ctx=None,
        )

    @property
    def errors(self) -> MappingProxyType[Any, ConstraintError]:
        # read-only access
        return MappingProxyType(self.args[1])

    def __repr__(self):
        return '{0}({{{1}}})'.format(
            self.__class__.__name__,
            ', '.join(f'{k!r}: {v!r}' for k, v in self.errors.items()),
        )


class ParameterContextErrors(Mapping):
    """Read-only convenience that wraps a ``ConstraintErrors`` error mapping
    """
    # TODO extend this class with any query functionality that a command
    # API would want to use in order to get streamlined information on what
    # went wrong (in general, for a specific parameter, etc...)
    def __init__(
        self,
        errors: Dict[ParameterConstraintContext, ConstraintError],
    ):
        self._errors = errors

    def __repr__(self):
        return self._errors.__repr__()

    def __len__(self):
        return len(self._errors)

    def __iter__(self):
        return self._errors.__iter__()

    def __getitem__(self, key):
        return self._errors[key]

    def items(self):
        return self._errors.items()

    @property
    def messages(self):
        return [e.msg for e in self._errors.values()]

    @property
    def context_labels(self):
        return [e.label for e in self._errors.keys()]

    # TODO return all errors related to some parameter


@dataclass(frozen=True)
class ParameterConstraintContext:
    """Representation of a parameter constraint context

    This type is used for the keys in the error map of.
    ``ParametrizationErrors``. Its purpose is to clearly identify which
    parameter combination (and its nature) led to a `ConstraintError`.

    An error context comprises to components: 1) the names of the parameters
    that were considered, and 2) a description of how the parameters were
    linked or combined. In the simple case of an error occurring in the context
    of a single parameter, the second component is superfluous. Otherwise,
    it can be thought of as an operation label, describing what aspect of
    the set of parameters is being relevant in a particular context.

    Example:

    A command has two parameters `p1` and `p2`. The may also have respective
    individual constraints, but importantly they 1) must not have identical
    values, and 2) their sum must be larger than 3. If the command is called
    with ``cmd(p1=1, p2=1)``, both conditions are violated. The reporting may
    be implemented using the following ``ParameterConstraintContext`` and
    ``ConstraintError`` instances::

      ParameterConstraintContext(('p1', 'p2'), 'inequality):
        ConstraintError(EnsureValue(True), False, <EnsureValue error>)

      ParameterConstraintContext(('p1', 'p2'), 'sum):
        ConstraintError(EnsureRange(min=3), False, <EnsureRange error>)

    where the ``ConstraintError`` instances are generated by standard
    ``Constraint`` implementation. For the second error, this could look like::

      EnsureRange(min=3)(params['p1'] + params['p2'])
    """
    parameters: Tuple[str]
    description: str | None = None

    def __str__(self):
        return f'Context<{self.label}>'

    @property
    def label(self) -> str:
        """A concise summary of the context

        This label will be a compact as possible.
        """
        # XXX this could be __str__ but its intended usage for rendering
        # a text description of all errors would seemingly forbid adding
        # type information -- which OTOH seems to be desirable for __str__
        return '{param}{descr}'.format(
            param=", ".join(self.parameters),
            descr=f" ({self.description})" if self.description else '',
        )

    def get_label_with_parameter_values(self, values: dict) -> str:
        """Like ``.label`` but each parameter will also state a value"""
        # TODO truncate the values after repr() to ensure a somewhat compact
        # output
        from .parameter import NoValue
        return '{param}{descr}'.format(
            param=", ".join(
                f'{p}=<no value>'
                if isinstance(values[p], NoValue)
                else f'{p}={values[p]!r}'
                for p in self.parameters
            ),
            descr=f" ({self.description})" if self.description else '',
        )


class ParametrizationErrors(ConstraintErrors):
    """Exception type raised on violating parameter constraints

    This is a ``ConstraintErrors`` variant that uses parameter names (i.e,
    ``str`` labels) as context identifiers. In addition to individual
    parameter names an additional ``__all__`` identifier is recognized. It
    can be used to record a ``ConstraintError`` arising from high-order
    constraints, such as the violation of "mutually exclusive" requirements
    across more than one parameter.
    """
    def __init__(
            self,
            exceptions: Dict[str, ConstraintError] |
                        Dict[ParameterConstraintContext, ConstraintError]):
        super().__init__(
            {k if isinstance(k, ParameterConstraintContext)
             else ParameterConstraintContext((k,)):
             v
             for k, v in exceptions.items()}
        )

    @property
    def errors(self) -> ParameterContextErrors:
        # read-only access
        return ParameterContextErrors(self.args[1])

    def __str__(self):
        return self._render_violations_as_indented_text_list(
            'parameter')

    def _render_violations_as_indented_text_list(self, violation_subject):
        violations = len(self.errors)

        return '{ne} {vs}constraint violation{p}\n{el}'.format(
            ne=violations,
            vs=f'{violation_subject} ' if violation_subject else '',
            p='s' if violations > 1 else '',
            el='\n'.join(
                '{ctx}\n{msg}'.format(
                    ctx=ctx.get_label_with_parameter_values(
                        c.value
                        if isinstance(c.value, dict)
                        else {ctx.parameters[0]: c.value}
                    ),
                    msg=indent(str(c), '  '),
                )
                for ctx, c in self.errors.items()
            ),
        )


class CommandParametrizationError(ParametrizationErrors):
    """Exception type raised on violating any command parameter constraints

    .. seealso::

       :mod:`~datalad_next.constraints.EnsureCommandParameterization`
    """
    def __str__(self):
        return self._render_violations_as_indented_text_list(
            'command parameter')



# ex: set sts=4 ts=4 sw=4 noet:
# -*- coding: utf-8 -*-
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""

"""

from pathlib import Path
from stat import S_IREAD, S_IRGRP, S_IROTH
from urllib.parse import quote as urlquote

from datalad.api import (
    Dataset,
    clone,
)
from datalad_next.tests import (
    DEFAULT_BRANCH,
    DEFAULT_REMOTE,
    assert_raises,
    assert_status,
    eq_,
    skip_if_root,
)
from datalad_next.consts import on_windows
from datalad_next.exceptions import CommandError
from datalad_next.utils import (
    patched_env,
    rmtree,
)
from ..datalad_annex import get_initremote_params_from_url


webdav_cred = ('datalad', 'secure')


def eq_dla_branch_state(state, path, branch=DEFAULT_BRANCH):
    """Confirm that the reported branch hexsha at a remote matches
    a given value"""
    refsfile = Path(path) / '3f7' / '4a3' / 'XDLRA--refs' / 'XDLRA--refs'
    if not refsfile.exists():
        # this may be an export remote
        refsfile = Path(path) / '.datalad' / 'dotgit' / 'refs'
    if not refsfile.exists():
        assert None, f'Could not find refs at {path}'

    for line in refsfile.read_text().splitlines():
        if line.strip().endswith(f'heads/{branch}'):
            eq_(state, line.split(maxsplit=1)[0])
            return
    assert None, f'Could not find state for branch {branch} at {path}'


@skip_if_root  # see https://github.com/datalad/datalad-next/issues/525
def test_annex_remote(existing_noannex_dataset, tmp_path, no_result_rendering):
    remotepath = tmp_path / 'remote'
    # bypass the complications of folding a windows path into a file URL
    dlaurl = \
        f'datalad-annex::?type=directory&directory={remotepath}&encryption=none' \
        if on_windows else \
        f'datalad-annex::file://{remotepath}?type=directory&directory={{path}}&encryption=none'
    ds = existing_noannex_dataset
    _check_push_fetch_cycle(ds, dlaurl, remotepath, tmp_path)


@skip_if_root  # see https://github.com/datalad/datalad-next/issues/525
def test_export_remote(existing_noannex_dataset, tmp_path, no_result_rendering):
    remotepath = tmp_path / 'remote'
    # bypass the complications of folding a windows path into a file URL
    dlaurl = \
        f'datalad-annex::?type=directory&directory={remotepath}&encryption=none&exporttree=yes' \
        if on_windows else \
        f'datalad-annex::file://{remotepath}?type=directory&directory={{path}}&encryption=none&exporttree=yes'
    ds = existing_noannex_dataset
    _check_push_fetch_cycle(ds, dlaurl, remotepath, tmp_path)


def _check_push_fetch_cycle(ds, remoteurl, remotepath, tmp_path):
    """Test helper

    - add a dla remote to the dataset
    - push the ds to it
    - clone from it to a tmp location
    - check error handling when post-git-update upload fails
    - update cycle starting from the original ds
    - repeated supposed-to-be-noop push/fetch calls
    - update cycle starting from the clone
    """
    localtargetpath = tmp_path / 'ltarget'
    probepath = tmp_path / 'probe'
    remotepath.mkdir()
    dsrepo = ds.repo
    dsrepo.call_git(['remote', 'add', 'dla', remoteurl])

    # basic push/clone roundtrip on clean locations
    # Since some version of git > 2.30.2 and <= 2.35.1
    # it would work without specifying branch.
    dsrepo.call_git(['push', '-u', 'dla', DEFAULT_BRANCH])
    eq_dla_branch_state(dsrepo.get_hexsha(DEFAULT_BRANCH), remotepath)
    dsclone = clone(remoteurl, localtargetpath)
    dsclonerepo = dsclone.repo
    eq_(dsrepo.get_hexsha(DEFAULT_BRANCH), dsclonerepo.get_hexsha(DEFAULT_BRANCH))

    # update round
    (ds.pathobj / 'file1').write_text('file1text')
    assert_status('ok', ds.save())

    # but first make destination read-only to test error recovery
    # verify starting point, we are one step ahead of the remote
    eq_(dsrepo.get_hexsha(DEFAULT_BRANCH + '~1'),
        dsrepo.get_hexsha(f'dla/{DEFAULT_BRANCH}'))

    # if we are on a sane system, also test recovery from (temporary)
    # push failure. MIH cannot force himself to figure out how to do
    # this on windows/crippledFS, sorry
    probeds = Dataset(probepath).create()
    if not probeds.repo.is_managed_branch():
        # preserve stat-info for later restore
        stat_records = {}
        # must go reverse to not block chmod'ing of children
        for p in sorted(remotepath.glob('**/*'), reverse=True):
            stat_records[p] = p.stat().st_mode
            p.chmod(S_IREAD | S_IRGRP | S_IROTH)
        # push must fail
        assert_raises(CommandError, dsrepo.call_git, ['push', 'dla'])
        # really bad that we cannot roll-back the remote branch state
        # from within the helper (see code), but we leave an indicator
        eq_(dsrepo.get_hexsha(DEFAULT_BRANCH),
            dsrepo.get_hexsha(f'refs/dlra-upload-failed/dla/{DEFAULT_BRANCH}'))

        # revert read-only permission on the remote side
        for p in sorted(stat_records):
            p.chmod(stat_records[p])

    # now a push can work (it should internally see that refs need
    # pushing that previously were reported as pushed, no need for
    # --force)
    dsrepo.call_git(['push', 'dla'])
    # and it has removed the marker
    assert_raises(
        ValueError,
        dsrepo.get_hexsha,
        f'refs/dlra-upload-failed/dla/{DEFAULT_BRANCH}')

    # the remote has received the new state
    eq_dla_branch_state(dsrepo.get_hexsha(DEFAULT_BRANCH), remotepath)
    # verify that there is something to update
    assert dsrepo.get_hexsha(DEFAULT_BRANCH) != dsclonerepo.get_hexsha(DEFAULT_BRANCH)
    # pull
    dsclonerepo.call_git(['pull', DEFAULT_REMOTE, DEFAULT_BRANCH])
    # source and clone are now equal
    eq_(dsrepo.get_hexsha(DEFAULT_BRANCH), dsclonerepo.get_hexsha(DEFAULT_BRANCH))

    # push no update
    dsrepo.call_git(['push', 'dla'])
    # twice
    dsrepo.call_git(['push', 'dla'])

    # fetch no update
    dsclonerepo.call_git(['fetch', DEFAULT_REMOTE])
    # twice
    dsclonerepo.call_git(['fetch', DEFAULT_REMOTE])

    # push/pull in reverse from clone to source
    (dsclone.pathobj / 'file2').write_text('file2text')
    assert_status('ok', dsclone.save())
    assert dsrepo.get_hexsha(DEFAULT_BRANCH) != dsclonerepo.get_hexsha(DEFAULT_BRANCH)
    dsclonerepo.call_git(['push', DEFAULT_REMOTE])
    eq_dla_branch_state(dsclonerepo.get_hexsha(DEFAULT_BRANCH), remotepath)
    dsrepo.call_git(['pull', 'dla', DEFAULT_BRANCH])
    eq_(dsrepo.get_hexsha(DEFAULT_BRANCH), dsclonerepo.get_hexsha(DEFAULT_BRANCH))

    # now create a non-heads ref and roundtrip that
    # this is what metalad needs to push metadata refs
    dsrepo.call_git([
        'update-ref', 'refs/datalad/dummy', dsrepo.get_hexsha(DEFAULT_BRANCH)])
    dsrepo.call_git(['push', 'dla', 'refs/datalad/dummy'])
    dsclonerepo.call_git([
        'fetch', DEFAULT_REMOTE, 'refs/datalad/dummy:refs/datalad/dummy'])
    eq_(dsrepo.get_hexsha('refs/datalad/dummy'),
        dsclonerepo.get_hexsha('refs/datalad/dummy'))


def test_annex_remote_autorepush(existing_noannex_dataset, tmp_path):
    remotepath = tmp_path
    # bypass the complications of folding a windows path into a file URL
    dlaurl = \
        f'datalad-annex::?type=directory&directory={remotepath}&encryption=none' \
        if on_windows else \
        f'datalad-annex::file://{remotepath}?type=directory&directory={{path}}&encryption=none'
    _check_repush_after_vanish(existing_noannex_dataset, dlaurl, remotepath)


def test_export_remote_autorepush(existing_noannex_dataset, tmp_path):
    remotepath = tmp_path
    # bypass the complications of folding a windows path into a file URL
    dlaurl = \
        f'datalad-annex::?type=directory&directory={remotepath}&encryption=none&exporttree=yes' \
        if on_windows else \
        f'datalad-annex::file://{remotepath}?type=directory&directory={{path}}&encryption=none&exporttree=yes'
    _check_repush_after_vanish(existing_noannex_dataset, dlaurl, remotepath)


def _check_repush_after_vanish(ds, remoteurl, remotepath):
    dsrepo = ds.repo
    dsrepo.call_git(['remote', 'add', 'dla', remoteurl])
    remotepath = Path(remotepath)

    dsrepo.call_git(['push', '-u', 'dla', DEFAULT_BRANCH])
    eq_dla_branch_state(dsrepo.get_hexsha(DEFAULT_BRANCH), remotepath)

    # wipe out the remote
    rmtree(remotepath)
    assert not remotepath.exists()
    remotepath.mkdir(parents=True)

    # helper must detect the discrepancy and re-push, despite the local mirror
    # repo already being up-to-date
    dsrepo.call_git(['push', 'dla'])
    eq_dla_branch_state(dsrepo.get_hexsha(DEFAULT_BRANCH), remotepath)


def test_params_from_url():
    f = get_initremote_params_from_url
    # just the query part being used
    eq_(f('datalad-annex::?encryption=none&type=directory&directory=/this/h'),
        ['encryption=none', 'type=directory', 'directory=/this/h'])
    # some url prperty expansion
    eq_(f('datalad-annex::file:///this/h?type=directory&directory={path}'),
        ['type=directory', 'directory=/this/h'])
    # original URL, but query stripped
    eq_(f('https://ex.com/dav/proj/ds?type=webdav&url={noquery}&keyid=id@ex'),
        ['type=webdav', 'url=https://ex.com/dav/proj/ds', 'keyid=id@ex'])
    # proper unquoting
    eq_(f('http://ex.com?weirdparam=some%26amp'),
        ['type=web', 'exporttree=yes',
         'url=http://ex.com?weirdparam=some%26amp'])
    # nothing is not valid
    assert_raises(ValueError, f, '')
    assert_raises(ValueError, f, 'datalad-annex::')
    # URL without annotation is type=web export remote
    eq_(f('http://example.com/path/to/something'),
        ['type=web', 'exporttree=yes',
         'url=http://example.com/path/to/something'])


def test_typeweb_annex(existing_noannex_dataset, http_server, tmp_path,
                       no_result_rendering):
    _check_typeweb(
        # bypass the complications of folding a windows path into a file URL
        'datalad-annex::?type=directory&directory={export}&encryption=none' \
        if on_windows else
        'datalad-annex::file://{export}?type=directory&directory={{path}}&encryption=none',
        'datalad-annex::{url}?type=web&url={{noquery}}',
        existing_noannex_dataset,
        http_server,
        tmp_path,
    )


# just to exercise the code path leading to an uncompressed ZIP
def test_typeweb_annex_uncompressed(
        existing_noannex_dataset, http_server, tmp_path, no_result_rendering):
    _check_typeweb(
        # bypass the complications of folding a windows path into a file URL
        'datalad-annex::?type=directory&directory={export}&encryption=none&dladotgit=uncompressed' \
        if on_windows else
        'datalad-annex::file://{export}?type=directory&directory={{path}}&encryption=none&dladotgit=uncompressed',
        'datalad-annex::{url}?type=web&url={{noquery}}',
        existing_noannex_dataset,
        http_server,
        tmp_path,
    )


def test_typeweb_export(existing_noannex_dataset, http_server, tmp_path,
                        no_result_rendering):
    _check_typeweb(
        # bypass the complications of folding a windows path into a file URL
        'datalad-annex::?type=directory&directory={export}&encryption=none&exporttree=yes' \
        if on_windows else
        'datalad-annex::file://{export}?type=directory&directory={{path}}&encryption=none&exporttree=yes',
        # when nothing is given type=web&exporttree=yes is the default
        'datalad-annex::{url}',
        existing_noannex_dataset,
        http_server,
        tmp_path,
    )


def _check_typeweb(pushtmpl, clonetmpl, ds, server, clonepath):
    ds.repo.call_git([
        'remote', 'add',
        'dla',
        pushtmpl.format(export=server.path),
    ])
    ds.repo.call_git(['push', '-u', 'dla', DEFAULT_BRANCH])
    # must override git-annex security setting for localhost
    with patched_env(**{
        "GIT_CONFIG_COUNT": "1",
        "GIT_CONFIG_KEY_0": "annex.security.allowed-ip-addresses",
        "GIT_CONFIG_VALUE_0": "127.0.0.1"}
    ):
        dsclone = clone(
            clonetmpl.format(url=server.url),
            clonepath)
    eq_(ds.repo.get_hexsha(DEFAULT_BRANCH),
        dsclone.repo.get_hexsha(DEFAULT_BRANCH))


def test_submodule_url(tmp_path, existing_noannex_dataset, http_server,
                       no_result_rendering):
    servepath = http_server.path
    url = http_server.url
    # a future subdataset that we want to register under a complex URL
    tobesubds = existing_noannex_dataset
    # push to test web server, this URL doesn't matter yet
    tobesubds.repo.call_git([
        'remote', 'add', 'dla',
        # bypass the complications of folding a windows path into a file URL
        f'datalad-annex::?type=directory&directory={servepath}&encryption=none&exporttree=yes'
        if on_windows else
        f'datalad-annex::file://{servepath}?type=directory&directory={{path}}&encryption=none&exporttree=yes',
    ])
    tobesubds.repo.call_git(['push', '-u', 'dla', DEFAULT_BRANCH])
    # create a superdataset to register the subds to
    super = Dataset(tmp_path / 'super').create()
    with patched_env(**{
        "GIT_CONFIG_COUNT": "1",
        "GIT_CONFIG_KEY_0": "annex.security.allowed-ip-addresses",
        "GIT_CONFIG_VALUE_0": "127.0.0.1"}
    ):
        # this is the URL that matters
        # we intentionally use something that leaves a placeholder behind
        # in the submodule record
        super.clone(
            f'datalad-annex::{url}?type=web&url={{noquery}}&exporttree=yes',
            'subds',
        )
    # no clone the entire super
    superclone = clone(super.path, tmp_path / 'superclone')
    # and auto-fetch the sub via the datalad-annex remote helper
    superclone.get('subds', get_data=False, recursive=True)
    # we got the original subds
    subdsclone = Dataset(superclone.pathobj / 'subds')
    eq_(tobesubds.id, subdsclone.id)


def test_webdav_auth(existing_noannex_dataset,
                     tmp_path,
                     credman,
                     webdav_credential,
                     webdav_server,
                     no_result_rendering):
    credman.set(**webdav_credential)
    # this is the dataset we want to roundtrip through webdav
    ds = existing_noannex_dataset
    remoteurl = \
        f'datalad-annex::{webdav_server.url}' \
        '?type=webdav&url={noquery}&encryption=none&' \
        f'dlacredential={urlquote(webdav_credential["name"])}'

    ds.repo.call_git(['remote', 'add', 'dla', remoteurl])

    # roundtrip
    ds.repo.call_git(['push', '-u', 'dla', DEFAULT_BRANCH])
    cln = clone(remoteurl, tmp_path)
    # must give the same thing
    eq_(ds.repo.get_hexsha(DEFAULT_BRANCH),
        cln.repo.get_hexsha(DEFAULT_BRANCH))



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Basic constraints for declaring essential data types, values, and ranges"""

from __future__ import annotations

__docformat__ = 'restructuredtext'

from hashlib import algorithms_guaranteed as hash_algorithms_guaranteed
from pathlib import Path
import re
from typing import Callable

from datalad_next.datasets import resolve_path

from .base import (
    Constraint,
    DatasetParameter,
)
from .utils import _type_str


class NoConstraint(Constraint):
    """A constraint that represents no constraints"""
    def short_description(self):
        return ''

    def __call__(self, value):
        return value


class EnsureValue(Constraint):
    """Ensure an input is a particular value"""
    def __init__(self, value):
        super().__init__()
        self._target_value = value

    def __call__(self, value):
        if value == self._target_value:
            return value
        else:
            self.raise_for(
                value,
                "must be {target_value!r}",
                target_value=self._target_value,
            )

    def short_description(self):
        return f'{self._target_value!r}'

    def long_description(self):
        return f'value must be {self.short_description()}'


class EnsureDType(Constraint):
    """Ensure that an input (or several inputs) are of a particular data type.
    Examples:

        >>> c = EnsureDType(float)
        >>> type(c(8))                  # doctest: +SKIP
        float
        >>> import numpy as np          # doctest: +SKIP
        >>> c = EnsureDType(np.float64) # doctest: +SKIP
        >>> type(c(8))                  # doctest: +SKIP
        numpy.float64
    """
    def __init__(self, dtype):
        """
        Parameters
        ----------
        dtype : functor
        """
        self._dtype = dtype

    def __call__(self, value):
        try:
            return self._dtype(value)
        except Exception as e:
            self.raise_for(
                value,
                str(e),
            )

    def short_description(self):
        return _type_str(self._dtype)

    def long_description(self):
        return "value must be convertible to type '%s'" % self.short_description()


class EnsureInt(EnsureDType):
    """Ensure that an input (or several inputs) are of a data type 'int'.
    """
    def __init__(self):
        """Initializes EnsureDType with int"""
        EnsureDType.__init__(self, int)


class EnsureFloat(EnsureDType):
    """Ensure that an input (or several inputs) are of a data type 'float'.
    """
    def __init__(self):
        """Initializes EnsureDType with float"""
        EnsureDType.__init__(self, float)


class EnsureBool(Constraint):
    """Ensure that an input is a bool.

    A couple of literal labels are supported, such as:
    False: '0', 'no', 'off', 'disable', 'false'
    True: '1', 'yes', 'on', 'enable', 'true'
    """
    def __call__(self, value):
        if isinstance(value, bool):
            return value
        elif isinstance(value, (bytes, str)):
            value = value.lower()
            if value in ('0', 'no', 'off', 'disable', 'false'):
                return False
            elif value in ('1', 'yes', 'on', 'enable', 'true'):
                return True
        self.raise_for(value, "must be convertible to boolean")

    def long_description(self):
        return 'value must be convertible to type bool'

    def short_description(self):
        return 'bool'


class EnsureStr(Constraint):
    """Ensure an input is a string of some min. length and matching a pattern

    Pattern matching is optional and minimum length is zero (empty string is
    OK).

    No type conversion is performed.
    """
    def __init__(self, min_len: int = 0, match: str | None = None):
        """
        Parameters
        ----------
        min_len: int, optional
           Minimal length for a string.
        match:
           Regular expression used to match any input value against.
           Values not matching the expression will cause a
           `ValueError` to be raised.
        """
        assert min_len >= 0
        self._min_len = min_len
        self._match = match
        super().__init__()
        if match is not None:
            self._match = re.compile(match)

    def __call__(self, value) -> str:
        if not isinstance(value, (bytes, str)):
            # do not perform a blind conversion ala str(), as almost
            # anything can be converted and the result is most likely
            # unintended
            self.raise_for(value, "must be a string")
        if len(value) < self._min_len:
            self.raise_for(value, "must have minimum length {len}",
                                 len=self._min_len)
        if self._match:
            if not self._match.match(value):
                self.raise_for(
                    value,
                    'does not match {pattern}',
                    pattern=self._match.pattern,
                )
        return value

    def long_description(self):
        return 'must be a string{}'.format(
            f' and match {self._match.pattern}' if self._match else '',
        )

    def short_description(self):
        return 'str{}'.format(
            f'({self._match.pattern})' if self._match else '',
        )


# TODO possibly consolidate on EnsureStr from -gooey, which can take
# a regex that could perform this. CON: documentation less clear.
# But if custom documentation will be supported, it might get even
# more clear nevertheless
class EnsureStrPrefix(EnsureStr):
    """Ensure an input is a string that starts with a given prefix.
    """
    def __init__(self, prefix):
        """
        Parameters
        ----------
        prefix : str
           Mandatory prefix.
        """
        self._prefix = prefix
        super().__init__()

    def __call__(self, value):
        super().__call__(value)
        if not value.startswith(self._prefix):
            self.raise_for(
                value,
                "does not start with {prefix!r}",
                prefix=self._prefix,
            )
        return value

    def long_description(self):
        return "value must start with '{}'".format(self._prefix)

    def short_description(self):
        return '{}...'.format(self._prefix)


class EnsureNone(EnsureValue):
    """Ensure an input is of value `None`"""
    def __init__(self):
        super().__init__(None)


class EnsureCallable(Constraint):
    """Ensure an input is a callable object"""
    def __call__(self, value):
        if hasattr(value, '__call__'):
            return value
        else:
            self.raise_for(value, "must be a callable")

    def short_description(self):
        return 'callable'

    def long_description(self):
        return 'value must be a callable'


class EnsureChoice(Constraint):
    """Ensure an input is element of a set of possible values"""

    def __init__(self, *values):
        """
        Parameters
        ----------
        *values
           Possible accepted values.
        """
        self._allowed = values
        super(EnsureChoice, self).__init__()

    def __call__(self, value):
        if value not in self._allowed:
            self.raise_for(
                value,
                "is not one of {allowed}",
                allowed=self._allowed,
            )
        return value

    def long_description(self):
        return 'value must be one of [CMD: %s CMD][PY: %s PY]' % (
            str(tuple(i for i in self._allowed if i is not None)),
            str(self._allowed)
        )

    def short_description(self):
        return '{%s}' % ', '.join([repr(c) for c in self._allowed])

    def __str__(self):
        return f"one of {self.short_description()}"


class EnsureKeyChoice(EnsureChoice):
    """Ensure value under a key in an input is in a set of possible values"""

    def __init__(self, key, values):
        """
        Parameters
        ----------
        key : str
          The to-be-tested values are looked up under the given key in
          a dict-like input object.
        values : tuple
           Possible accepted values.
        """
        self._key = key
        super(EnsureKeyChoice, self).__init__(*values)

    def __call__(self, value):
        if self._key not in value:
            self.raise_for(value, "must be dict-like")
        super(EnsureKeyChoice, self).__call__(value[self._key])
        return value

    def long_description(self):
        return "value in '%s' must be one of %s" % (self._key, str(self._allowed),)

    def short_description(self):
        return '%s:{%s}' % (self._key, ', '.join([repr(c) for c in self._allowed]))


class EnsureRange(Constraint):
    """Ensure an input is within a particular range

    No type checks are performed.
    """
    def __init__(self, min=None, max=None):
        """
        Parameters
        ----------
        min
            Minimal value to be accepted in the range
        max
            Maximal value to be accepted in the range
        """
        self._min = min
        self._max = max
        if self._min is None and self._max is None:
            raise ValueError('No range given, min == max == None')
        super(EnsureRange, self).__init__()

    def __call__(self, value):
        if self._min is not None:
            if self._max is not None:
                if value < self._min or value > self._max:
                    self.raise_for(
                        value,
                        f"must be in range from {self._min!r} to {self._max!r}"
                    )
            else:
                if value < self._min:
                    self.raise_for(value, f"must be at least {self._min!r}")
        if self._max is not None:
            if value > self._max:
                self.raise_for(value, f"must be at most {self._max!r}")
        return value

    def long_description(self):
        return self.short_description()

    def short_description(self):
        if self._max is None:
            return f'not less than {self._min!r}'
        elif self._min is None:
            return f'not greater than {self._max!r}'
        else:
            # it is inclusive, but spelling it out would be wordy
            return f'in range from {self._min!r} to {self._max!r}'


class EnsurePath(Constraint):
    """Ensures input is convertible to a (platform) path and returns a `Path`

    Optionally, the path can be tested for existence and whether it is absolute
    or relative.
    """
    def __init__(self,
                 *,
                 path_type: type = Path,
                 is_format: str | None = None,
                 lexists: bool | None = None,
                 is_mode: Callable | None = None,
                 ref: Path | None = None,
                 ref_is: str = 'parent-or-same-as',
                 dsarg: DatasetParameter | None = None):
        """
        Parameters
        ----------
        path_type:
          Specific pathlib type to convert the input to. The default is `Path`,
          i.e. the platform's path type. Not all pathlib Path types can be
          instantiated on all platforms, and not all checks are possible with
          all path types.
        is_format: {'absolute', 'relative'} or None
          If not None, the path is tested whether it matches being relative or
          absolute.
        lexists:
          If not None, the path is tested to confirmed exists or not. A symlink
          need not point to an existing path to fulfil the "exists" condition.
        is_mode:
          If set, this callable will receive the path's `.lstat().st_mode`,
          and an exception is raised, if the return value does not evaluate
          to `True`. Typical callables for this feature are provided by the
          `stat` module, e.g. `S_ISDIR()`
        ref:
          If set, defines a reference Path any given path is compared to. The
          comparison operation is given by `ref_is`.
        ref_is: {'parent-or-same-as', 'parent-of'}
          Comparison operation to perform when `ref` is given.
        dsarg: DatasetParameter, optional
          If given, incoming paths are resolved in the following fashion:
          If, and only if, the original "dataset" parameter was a
          ``Dataset`` object instance, relative paths are interpreted as
          relative to the given dataset. In all other cases, relative paths
          are treated as relative to the current working directory.
        """
        super().__init__()
        self._path_type = path_type
        self._is_format = is_format
        self._lexists = lexists
        self._is_mode = is_mode
        self._ref = ref
        self._ref_is = ref_is
        self._dsarg = dsarg
        assert self._ref_is in ('parent-or-same-as', 'parent-of'), \
            'Unrecognized `ref_is` operation label'

    def __call__(self, value):
        # turn it into the target type to make everything below
        # more straightforward
        path = self._path_type(value)

        # we are testing the format first, because resolve_path()
        # will always turn things into absolute paths
        if self._is_format is not None:
            is_abs = path.is_absolute()
            if self._is_format == 'absolute' and not is_abs:
                self.raise_for(path, 'is not an absolute path')
            elif self._is_format == 'relative' and is_abs:
                self.raise_for(path, 'is not a relative path')

        # resolve relative paths against a dataset, if given
        if self._dsarg:
            path = resolve_path(
                path,
                self._dsarg.original,
                self._dsarg.ds)

        mode = None
        if self._lexists is not None or self._is_mode is not None:
            try:
                mode = path.lstat().st_mode
            except FileNotFoundError:
                # this is fine, handled below
                pass
        if self._lexists is not None:
            if self._lexists and mode is None:
                self.raise_for(path, 'does not exist')
            elif not self._lexists and mode is not None:
                self.raise_for(path, 'does (already) exist')
        if self._is_mode is not None:
            if not self._is_mode(mode):
                self.raise_for(path, 'does not match desired mode')
        if self._ref:
            ok = True
            if self._ref_is == 'parent-or-same-as':
                ok = (path == self._ref or self._ref in path.parents)
            elif self._ref_is == 'parent-of':
                ok = self._ref in path.parents
            else:  # pragma: nocover
                # this code cannot be reached with normal usage.
                # it is prevented by an assertion in __init__()
                raise RuntimeError('Unknown `ref_is` operation label')

            if not ok:
                self.raise_for(
                    path,
                    '{ref} is not {ref_is} {path}',
                    ref=self._ref,
                    ref_is=self._ref_is,
                )
        return path

    def for_dataset(self, dataset: DatasetParameter) -> Constraint:
        """Return an similarly parametrized variant that resolves
        paths against a given dataset (argument)


        """
        return self.__class__(
            path_type=self._path_type,
            is_format=self._is_format,
            lexists=self._lexists,
            is_mode=self._is_mode,
            ref=self._ref,
            ref_is=self._ref_is,
            dsarg=dataset,
        )

    def short_description(self):
        return '{}{}path{}'.format(
            'existing '
            if self._lexists
            else 'non-existing '
            if self._lexists else '',
            'absolute '
            if self._is_format == 'absolute'
            else 'relative'
            if self._is_format == 'relative'
            else '',
            f' that is {self._ref_is} {self._ref}'
            if self._ref
            else '',
        )


class EnsureHashAlgorithm(EnsureChoice):
    """Ensure an input matches a name of a ``hashlib`` algorithm

    Specifically the item must be in the ``algorithms_guaranteed`` collection.
    """
    def __init__(self):
        super().__init__(*hash_algorithms_guaranteed)
